import argparse
import base64
import sys
from typing import Dict, Iterable, List, Set, Text, Tuple, Union, Any

import questionary
import requests
import structlog

import rasa.cli.telemetry
import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.shared.constants import (
    DEFAULT_DOMAIN_PATHS,
    DEFAULT_CONFIG_PATH,
)
from rasa.shared.core.flows.yaml_flows_io import YamlFlowsWriter
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter, FlowSyncImporter
from rasa.shared.nlu.training_data.formats.rasa_yaml import RasaYAMLWriter
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string, read_yaml_file
from rasa.studio import results_logger
from rasa.studio.auth import KeycloakTokenReader
from rasa.studio.config import StudioConfig
from rasa.studio.results_logger import StudioResult, with_studio_error_handler

structlogger = structlog.get_logger()


def _get_selected_entities_and_intents(
    args: argparse.Namespace,
    intents_from_files: Set[Text],
    entities_from_files: List[Text],
) -> Tuple[List[Text], List[Text]]:
    entities = args.entities

    if entities is None or len(entities) == 0:
        entities = entities_from_files
        structlogger.info(
            "rasa.studio.upload.entities_empty",
            event_info="No entities specified. Using all entities from files.",
        )

    intents = args.intents

    if intents is None or len(intents) == 0:
        intents = intents_from_files
        structlogger.info(
            "rasa.studio.upload.intents_empty",
            event_info="No intents specified. Using all intents from files.",
        )

    return list(entities), list(intents)


def handle_upload(args: argparse.Namespace) -> None:
    """Uploads primitives to rasa studio."""
    studio_config = StudioConfig.read_config()
    endpoint = studio_config.studio_url
    verify = not studio_config.disable_verify

    if not endpoint:
        rasa.shared.utils.cli.print_error_and_exit(
            "No GraphQL endpoint found in config. Please run `rasa studio config`."
        )
    else:
        structlogger.info(
            "rasa.studio.upload.loading_data", event_info="Loading data..."
        )

        args.domain = rasa.cli.utils.get_validated_path(
            args.domain, "domain", DEFAULT_DOMAIN_PATHS
        )

        args.config = rasa.cli.utils.get_validated_path(
            args.config, "config", DEFAULT_CONFIG_PATH
        )

        # check safely if args.calm is set and not fail if not
        if hasattr(args, "calm") and args.calm:
            upload_calm_assistant(args, endpoint, verify=verify)
        else:
            upload_nlu_assistant(args, endpoint, verify=verify)


config_keys = [
    "recipe",
    "language",
    "pipeline",
    "llm",
    "policies",
    "model_name",
    "assistant_id",
]


def extract_values(data: Dict, keys: List[Text]) -> Dict:
    """Extracts values for given keys from a dictionary."""
    return {key: data.get(key) for key in keys if data.get(key)}


def _get_assistant_name(config: Dict[Text, Any]) -> str:
    config_assistant_id = config.get("assistant_id", "")
    assistant_name = questionary.text(
        "Please provide the assistant name", default=config_assistant_id
    ).ask()
    if not assistant_name:
        structlogger.error(
            "rasa.studio.upload.assistant_name_empty",
            event_info="Assistant name cannot be empty.",
        )
        sys.exit(1)

    # if assistant_name exists and different from config assistant_id,
    # notify user and upload with new assistant_name
    same = assistant_name == config_assistant_id
    if not same and config_assistant_id != "":
        structlogger.info(
            "rasa.studio.upload.assistant_name_mismatch",
            event_info=(
                f"Assistant name '{assistant_name}' is different"
                f" from the one in the config file: '{config_assistant_id}'."
            ),
        )

    structlogger.info(f"Uploading assistant with the name '{assistant_name}'.")
    return assistant_name


@with_studio_error_handler
def upload_calm_assistant(
    args: argparse.Namespace, endpoint: str, verify: bool = True
) -> StudioResult:
    """Uploads the CALM assistant data to Rasa Studio.

    Args:
        args: The command line arguments
            - data: The path to the training data
            - domain: The path to the domain
            - flows: The path to the flows
            - endpoints: The path to the endpoints
            - config: The path to the config
        endpoint: The studio endpoint
    Returns:
        None
    """
    structlogger.info(
        "rasa.studio.upload.loading_data", event_info="Parsing CALM assistant data..."
    )

    importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain,
        config_path=args.config,
    )

    # Prepare config and domain
    config = importer.get_config()
    domain_from_files = importer.get_user_domain().as_dict()
    endpoints_from_files = read_yaml_file(args.endpoints)
    config_from_files = read_yaml_file(args.config)

    # Extract domain and config values
    domain_keys = [
        "version",
        "actions",
        "responses",
        "slots",
        "intents",
        "entities",
        "forms",
        "session_config",
    ]

    domain = extract_values(domain_from_files, domain_keys)

    assistant_name = _get_assistant_name(config)

    training_data_paths = args.data

    if isinstance(training_data_paths, list):
        training_data_paths.append(args.flows)
    elif isinstance(training_data_paths, str):
        if isinstance(args.flows, list):
            training_data_paths = [training_data_paths] + args.flows
        elif isinstance(args.flows, str):
            training_data_paths = [training_data_paths, args.flows]
        else:
            raise RasaException("Invalid flows path")

    # Prepare flows
    flow_importer = FlowSyncImporter.load_from_dict(
        training_data_paths=training_data_paths
    )

    flows = list(flow_importer.get_user_flows())

    # We instantiate the TrainingDataImporter again on purpose to avoid
    # adding patterns to domain's actions. More info https://t.ly/W8uuc
    nlu_importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain, training_data_paths=args.data
    )
    nlu_data = nlu_importer.get_nlu_data()

    intents_from_files = nlu_data.intents

    nlu_examples = nlu_data.filter_training_examples(
        lambda ex: ex.get("intent") in intents_from_files
    )

    nlu_examples_yaml = RasaYAMLWriter().dumps(nlu_examples)

    # Build GraphQL request
    graphql_req = build_import_request(
        assistant_name,
        flows_yaml=YamlFlowsWriter().dumps(flows),
        domain_yaml=dump_obj_as_yaml_to_string(domain),
        config_yaml=dump_obj_as_yaml_to_string(config_from_files),
        endpoints=dump_obj_as_yaml_to_string(endpoints_from_files),
        nlu_yaml=nlu_examples_yaml,
    )

    structlogger.info("Uploading to Rasa Studio...")
    return make_request(endpoint, graphql_req, verify)


@with_studio_error_handler
def upload_nlu_assistant(
    args: argparse.Namespace, endpoint: str, verify: bool = True
) -> StudioResult:
    """Uploads the classic (dm1) assistant data to Rasa Studio.

    Args:
        args: The command line arguments
            - data: The path to the training data
            - domain: The path to the domain
            - intents: The intents to upload
            - entities: The entities to upload
        endpoint: The studio endpoint
    Returns:
        None
    """
    structlogger.info("Found DM1 assistant data, parsing...")
    importer = TrainingDataImporter.load_from_dict(
        domain_path=args.domain, training_data_paths=args.data, config_path=args.config
    )

    intents_from_files = importer.get_nlu_data().intents
    entities_from_files = importer.get_domain().entities

    entities, intents = _get_selected_entities_and_intents(
        args, intents_from_files, entities_from_files
    )

    config_from_files = importer.get_config()
    config = extract_values(config_from_files, config_keys)

    assistant_name = _get_assistant_name(config)

    structlogger.info("Validating data...")
    _check_for_missing_primitives(
        intents, entities, intents_from_files, entities_from_files
    )

    nlu_examples = importer.get_nlu_data().filter_training_examples(
        lambda ex: ex.get("intent") in intents
    )

    all_entities = _add_missing_entities(nlu_examples.entities, entities)
    nlu_examples_yaml = RasaYAMLWriter().dumps(nlu_examples)

    domain = _filter_domain(all_entities, intents, importer.get_domain().as_dict())
    domain_yaml = dump_obj_as_yaml_to_string(domain)

    graphql_req = build_request(assistant_name, nlu_examples_yaml, domain_yaml)

    structlogger.info("Uploading to Rasa Studio...")
    return make_request(endpoint, graphql_req, verify)


def make_request(endpoint: str, graphql_req: Dict, verify: bool = True) -> StudioResult:
    """Makes a request to the studio endpoint to upload data.

    Args:
        endpoint: The studio endpoint
        graphql_req: The graphql request
        verify: Whether to verify SSL
    """
    token = KeycloakTokenReader().get_token()
    res = requests.post(
        endpoint,
        json=graphql_req,
        headers={
            "Authorization": f"{token.token_type} {token.access_token}",
            "Content-Type": "application/json",
        },
        verify=verify,
    )

    if results_logger.response_has_errors(res.json()):
        return StudioResult.error(res.json())
    return StudioResult.success("Upload successful")


def _add_missing_entities(
    entities_from_intents: Iterable[str], entities: List[str]
) -> List[Union[str, Dict]]:
    all_entities: List[Union[str, Dict]] = []
    all_entities.extend(entities)
    for entity in entities_from_intents:
        if entity not in entities:
            structlogger.warning(
                f"Adding entity '{entity}' to upload since it is used in an intent."
            )
            all_entities.append(entity)
    return all_entities


def build_import_request(
    assistant_name: str,
    flows_yaml: str,
    domain_yaml: str,
    config_yaml: str,
    endpoints: str,
    nlu_yaml: str = "",
) -> Dict:
    # b64encode expects bytes and returns bytes so we need to decode to string
    base64_domain = base64.b64encode(domain_yaml.encode("utf-8")).decode("utf-8")
    base64_flows = base64.b64encode(flows_yaml.encode("utf-8")).decode("utf-8")
    base64_config = base64.b64encode(config_yaml.encode("utf-8")).decode("utf-8")
    base64_nlu = base64.b64encode(nlu_yaml.encode("utf-8")).decode("utf-8")
    base64_endpoints = base64.b64encode(endpoints.encode("utf-8")).decode("utf-8")

    graphql_req = {
        "query": (
            "mutation UploadModernAssistant($input: UploadModernAssistantInput!)"
            "{\n  uploadModernAssistant(input: $input)\n}"
        ),
        "variables": {
            "input": {
                "assistantName": assistant_name,
                "domain": base64_domain,
                "flows": base64_flows,
                "nlu": base64_nlu,
                "config": base64_config,
                "endpoints": base64_endpoints,
            }
        },
    }

    return graphql_req


def build_request(
    assistant_name: str, nlu_examples_yaml: str, domain_yaml: str
) -> Dict:
    # b64encode expects bytes and returns bytes so we need to decode to string
    base64_domain = base64.b64encode(domain_yaml.encode("utf-8")).decode("utf-8")
    base64_nlu_examples = base64.b64encode(nlu_examples_yaml.encode("utf-8")).decode(
        "utf-8"
    )

    graphql_req = {
        "query": (
            "mutation ImportFromEncodedYaml($input: ImportFromEncodedYamlInput!)"
            "{\n  importFromEncodedYaml(input: $input)\n}"
        ),
        "variables": {
            "input": {
                "assistantName": assistant_name,
                "domain": base64_domain,
                "nlu": base64_nlu_examples,
            }
        },
    }

    return graphql_req


def _filter_domain(
    entities: List[Union[str, Dict]], intents: List[str], domain_from_files: Dict
) -> Dict:
    """Filters the domain to only include the selected entities and intents."""
    selected_entities = _remove_not_selected_entities(
        entities, domain_from_files.get("entities", [])
    )
    return {
        "version": domain_from_files["version"],
        "intents": intents,
        "entities": selected_entities,
    }


def _check_for_missing_primitives(
    intents: Iterable[str],
    entities: Iterable[str],
    intents_found: Iterable[str],
    entities_found: Iterable[str],
) -> None:
    """Checks if the data contains all intents and entities.

    Args:
        intents: Iterable of intents to check
        entities: Iterable of entities to check
        intents_found: Iterable of intents found in the data
        entities_found: Iterable of entities found in the data
    """
    not_found_entities = [entity for entity in entities if entity not in entities_found]
    not_found_intents = [intent for intent in intents if intent not in intents_found]

    not_found_msg = ""
    if not_found_entities:
        not_found_msg += (
            f"Entities were not found in the training data: {not_found_entities}\n"
        )
    if not_found_intents:
        not_found_msg += (
            f"Intents were not found in the training data: {not_found_intents}\n"
        )

    if not_found_msg:
        raise RasaException(not_found_msg)


def _remove_not_selected_entities(
    entities: List[Union[Text, Dict]], domain_entities: List[Union[Text, Dict]]
) -> List:
    to_remove: List[Union[Text, Dict]] = []

    for entity in domain_entities:
        if isinstance(entity, str) and entity not in entities:
            to_remove.append(entity)
        elif isinstance(entity, dict) and len(entity) == 1:
            entity_name = next(iter(entity))
            if entity_name not in entities:
                to_remove.append(entity)

    for entity in to_remove:
        domain_entities.remove(entity)

    return domain_entities



import logging
from typing import Any, Dict, Text, Type

from rasa.engine.caching import TrainingCache
from rasa.engine.graph import ExecutionContext, GraphNodeHook, GraphSchema, SchemaNode
from rasa.engine.storage.storage import ModelStorage
from rasa.engine.training.components import PrecomputedValueProvider
import rasa.shared.utils.io
from rasa.engine.training import fingerprinting

logger = logging.getLogger(__name__)


class TrainingHook(GraphNodeHook):
    """Caches fingerprints and outputs of nodes during model training."""

    def __init__(
        self,
        cache: TrainingCache,
        model_storage: ModelStorage,
        pruned_schema: GraphSchema,
    ) -> None:
        """Initializes a `TrainingHook`.

        Args:
            cache: Cache used to store fingerprints and outputs.
            model_storage: Used to cache `Resource`s.
            pruned_schema: The pruned training schema.
        """
        self._cache = cache
        self._model_storage = model_storage
        self._pruned_schema = pruned_schema

    def on_before_node(
        self,
        node_name: Text,
        execution_context: ExecutionContext,
        config: Dict[Text, Any],
        received_inputs: Dict[Text, Any],
    ) -> Dict:
        """Calculates the run fingerprint for use in `on_after_node`."""
        graph_component_class = self._get_graph_component_class(
            execution_context, node_name
        )
        graph_component_config = self._get_graph_component_config(
            execution_context, node_name
        )

        merged_config = {
            **config,
            **graph_component_config,
        }

        fingerprint_key = fingerprinting.calculate_fingerprint_key(
            graph_component_class=graph_component_class,
            config=merged_config,
            inputs=received_inputs,
        )

        return {"fingerprint_key": fingerprint_key}

    def on_after_node(
        self,
        node_name: Text,
        execution_context: ExecutionContext,
        config: Dict[Text, Any],
        output: Any,
        input_hook_data: Dict,
    ) -> None:
        """Stores the fingerprints and caches the output of the node."""
        # We should not re-cache the output of a PrecomputedValueProvider.
        graph_component_class = self._pruned_schema.nodes[node_name].uses

        if graph_component_class == PrecomputedValueProvider:
            return None

        output_fingerprint = rasa.shared.utils.io.deep_container_fingerprint(output)
        fingerprint_key = input_hook_data["fingerprint_key"]

        logger.debug(
            f"Caching '{output.__class__.__name__}' with fingerprint_key: "
            f"'{fingerprint_key}' and output_fingerprint '{output_fingerprint}'."
        )

        self._cache.cache_output(
            fingerprint_key=fingerprint_key,
            output=output,
            output_fingerprint=output_fingerprint,
            model_storage=self._model_storage,
        )

    @staticmethod
    def _get_graph_component_class(
        execution_context: ExecutionContext, node_name: Text
    ) -> Type:
        graph_component_class = execution_context.graph_schema.nodes[node_name].uses
        return graph_component_class

    @staticmethod
    def _get_graph_component_config(
        execution_context: ExecutionContext, node_name: str
    ) -> Dict[str, Any]:
        return execution_context.graph_schema.nodes[node_name].config


class LoggingHook(GraphNodeHook):
    """Logs the training of components."""

    def __init__(self, pruned_schema: GraphSchema) -> None:
        """Creates hook.

        Args:
            pruned_schema: The pruned schema provides us with the information whether
                a component is cached or not.
        """
        self._pruned_schema = pruned_schema

    def on_before_node(
        self,
        node_name: Text,
        execution_context: ExecutionContext,
        config: Dict[Text, Any],
        received_inputs: Dict[Text, Any],
    ) -> Dict:
        """Logs the training start of a graph node."""
        node = self._pruned_schema.nodes[node_name]

        if not self._is_cached_node(node) and self._does_node_train(node):
            logger.info(f"Starting to train component '{node.uses.__name__}'.")

        return {}

    @staticmethod
    def _does_node_train(node: SchemaNode) -> bool:
        # Nodes which train are always targets so that they store their output in the
        # model storage. `is_input` filters out nodes which don't really train but e.g.
        # persist some training data.
        return node.is_target and not node.is_input

    @staticmethod
    def _is_cached_node(node: SchemaNode) -> bool:
        return node.uses == PrecomputedValueProvider

    def on_after_node(
        self,
        node_name: Text,
        execution_context: ExecutionContext,
        config: Dict[Text, Any],
        output: Any,
        input_hook_data: Dict,
    ) -> None:
        """Logs when a component finished its training."""
        node = self._pruned_schema.nodes[node_name]

        if not self._does_node_train(node):
            return

        if self._is_cached_node(node):
            actual_component = execution_context.graph_schema.nodes[node_name]
            logger.info(
                f"Restored component '{actual_component.uses.__name__}' from cache."
            )
        else:
            logger.info(f"Finished training component '{node.uses.__name__}'.")



from typing import Any, Optional, Tuple, Text, Dict, Set, List

import typing
import copy

import rasa.shared.utils.io
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import (
    COMMANDS,
    TEXT,
    INTENT,
    RESPONSE,
    INTENT_RESPONSE_KEY,
    METADATA,
    METADATA_INTENT,
    METADATA_EXAMPLE,
    ENTITIES,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_END,
    RESPONSE_IDENTIFIER_DELIMITER,
    FEATURE_TYPE_SENTENCE,
    FEATURE_TYPE_SEQUENCE,
    ACTION_TEXT,
    ACTION_NAME,
    TEXT_TOKENS,
)
from rasa.shared.constants import DIAGNOSTIC_DATA

if typing.TYPE_CHECKING:
    from rasa.shared.nlu.training_data.features import Features


class Message:
    """Container for data that can be used to describe a conversation turn.

    The turn is described by a set of attributes such as e.g. `TEXT`  and  `INTENT`
    when describing a user utterance or e.g. `ACTION_NAME` for describing a bot action.
    The container includes raw information (`self.data`) as well as features
    (`self.features`) for each such attribute.
    Moreover, the message has a timestamp and can keep track about information
    on a specific subset of attributes (`self.output_properties`).
    """

    def __init__(
        self,
        data: Optional[Dict[Text, Any]] = None,
        output_properties: Optional[Set] = None,
        time: Optional[int] = None,
        features: Optional[List["Features"]] = None,
        **kwargs: Any,
    ) -> None:
        """Creates an instance of Message."""
        self.time = time
        self.data = data.copy() if data else {}
        self.features = features if features else []

        self.data.update(**kwargs)
        self._cached_fingerprint: Optional[Text] = None

        if output_properties:
            self.output_properties = output_properties
        else:
            self.output_properties = set()
        self.output_properties.add(TEXT)

    def add_features(self, features: Optional["Features"]) -> None:
        """Add more vectorized features to the message."""
        if features is not None:
            self.features.append(features)
        self._cached_fingerprint = None

    def add_diagnostic_data(self, origin: Text, data: Dict[Text, Any]) -> None:
        """Adds diagnostic data from the `origin` component.

        Args:
            origin: Name of the component that created the data.
            data: The diagnostic data.
        """
        if origin in self.get(DIAGNOSTIC_DATA, {}):
            rasa.shared.utils.io.raise_warning(
                f"Please make sure every pipeline component has a distinct name. "
                f"The name '{origin}' appears at least twice and diagnostic "
                f"data will be overwritten."
            )
        self.data.setdefault(DIAGNOSTIC_DATA, {})
        self.data[DIAGNOSTIC_DATA][origin] = data
        self._cached_fingerprint = None

    def set(self, prop: Text, info: Any, add_to_output: bool = False) -> None:
        """Sets the message's property to the given value.

        Args:
            prop: Name of the property to be set.
            info: Value to be assigned to that property.
            add_to_output: Decides whether to add `prop` to the `output_properties`.
        """
        self.data[prop] = info
        if add_to_output:
            self.output_properties.add(prop)
        self._cached_fingerprint = None

    def get(self, prop: Text, default: Optional[Any] = None) -> Any:
        """Retrieve message property."""
        return self.data.get(prop, default)

    def as_dict_nlu(self) -> dict:
        """Get dict representation of message as it would appear in training data."""
        d = self.as_dict()
        if d.get(INTENT, None):
            d[INTENT] = self.get_full_intent()
        d.pop(RESPONSE, None)
        d.pop(INTENT_RESPONSE_KEY, None)
        return d

    def as_dict(self, only_output_properties: bool = False) -> Dict:
        """Gets dict representation of message."""
        if only_output_properties:
            d = {}
            for key, value in self.data.items():
                if key in self.output_properties:
                    if key == TEXT_TOKENS:
                        d[TEXT_TOKENS] = [(t.start, t.end) for t in value]
                    else:
                        d[key] = value
        else:
            d = self.data

        # Filter all keys with None value. These could have come while building the
        # Message object in markdown format
        return {key: value for key, value in d.items() if value is not None}

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Message):
            return False
        else:
            return other.fingerprint() == self.fingerprint()

    def __hash__(self) -> int:
        """Calculate a hash for the message.

        Returns:
            Hash of the message.
        """
        return int(self.fingerprint(), 16)

    def fingerprint(self) -> Text:
        """Calculate a string fingerprint for the message.

        Returns:
            Fingerprint of the message.
        """
        if self._cached_fingerprint is None:
            self._cached_fingerprint = rasa.shared.utils.io.deep_container_fingerprint(
                [self.data, self.features]
            )
        return self._cached_fingerprint

    @classmethod
    def build(
        cls,
        text: Text,
        intent: Optional[Text] = None,
        entities: Optional[List[Dict[Text, Any]]] = None,
        intent_metadata: Optional[Any] = None,
        example_metadata: Optional[Any] = None,
        **kwargs: Any,
    ) -> "Message":
        """Builds a Message from `UserUttered` data.

        Args:
            text: text of a user's utterance
            intent: an intent of the user utterance
            entities: entities in the user's utterance
            intent_metadata: optional metadata for the intent
            example_metadata: optional metadata for the intent example

        Returns:
            Message
        """
        data: Dict[Text, Any] = {TEXT: text}
        if intent:
            split_intent, response_key = cls.separate_intent_response_key(intent)
            if split_intent:
                data[INTENT] = split_intent
            if response_key:
                # intent label can be of the form - {intent}/{response_key},
                # so store the full intent label in intent_response_key
                data[INTENT_RESPONSE_KEY] = intent
        if entities:
            data[ENTITIES] = entities
        if intent_metadata is not None:
            data[METADATA] = {METADATA_INTENT: intent_metadata}
        if example_metadata is not None:
            data.setdefault(METADATA, {})[METADATA_EXAMPLE] = example_metadata

        return cls(data, **kwargs)

    def get_full_intent(self) -> Text:
        """Get intent as it appears in training data."""
        return (
            self.get(INTENT_RESPONSE_KEY)
            if self.get(INTENT_RESPONSE_KEY)
            else self.get(INTENT)
        )

    @staticmethod
    def separate_intent_response_key(
        original_intent: Text,
    ) -> Tuple[Text, Optional[Text]]:
        """Splits intent into main intent name and optional sub-intent name.

        For example, `"FAQ/how_to_contribute"` would be split into
        `("FAQ", "how_to_contribute")`. The response delimiter can
        take different values (not just `"/"`) and depends on the
        constant - `RESPONSE_IDENTIFIER_DELIMITER`.
        If there is no response delimiter in the intent, the second tuple
        item is `None`, e.g. `"FAQ"` would be mapped to `("FAQ", None)`.
        """
        split_title = original_intent.split(RESPONSE_IDENTIFIER_DELIMITER)
        if len(split_title) == 2:
            return split_title[0], split_title[1]
        elif len(split_title) == 1:
            return split_title[0], None

        raise RasaException(
            f"Intent name '{original_intent}' is invalid, "
            f"it cannot contain more than one '{RESPONSE_IDENTIFIER_DELIMITER}'."
        )

    def get_sparse_features(
        self, attribute: Text, featurizers: Optional[List[Text]] = None
    ) -> Tuple[Optional["Features"], Optional["Features"]]:
        """Gets all sparse features for the attribute given the list of featurizers.

        If no featurizers are provided, all available features will be considered.

        Args:
            attribute: message attribute
            featurizers: names of featurizers to consider

        Returns:
            Sparse features.
        """
        if featurizers is None:
            featurizers = []

        sequence_features, sentence_features = self._filter_sparse_features(
            attribute, featurizers
        )

        combined_sequence_features = self._combine_features(
            sequence_features, featurizers
        )
        combined_sentence_features = self._combine_features(
            sentence_features, featurizers
        )

        return combined_sequence_features, combined_sentence_features

    def get_sparse_feature_sizes(
        self, attribute: Text, featurizers: Optional[List[Text]] = None
    ) -> Dict[Text, List[int]]:
        """Gets sparse feature sizes for the attribute given the list of featurizers.

        If no featurizers are provided, all available features will be considered.

        Args:
            attribute: message attribute
            featurizers: names of featurizers to consider

        Returns:
            Sparse feature sizes.
        """
        if featurizers is None:
            featurizers = []

        sequence_features, sentence_features = self._filter_sparse_features(
            attribute, featurizers
        )
        sequence_sizes = [f.features.shape[1] for f in sequence_features]
        sentence_sizes = [f.features.shape[1] for f in sentence_features]

        return {
            FEATURE_TYPE_SEQUENCE: sequence_sizes,
            FEATURE_TYPE_SENTENCE: sentence_sizes,
        }

    def get_dense_features(
        self, attribute: Text, featurizers: Optional[List[Text]] = None
    ) -> Tuple[Optional["Features"], Optional["Features"]]:
        """Gets all dense features for the attribute given the list of featurizers.

        If no featurizers are provided, all available features will be considered.

        Args:
            attribute: message attribute
            featurizers: names of featurizers to consider

        Returns:
            Dense features.
        """
        if featurizers is None:
            featurizers = []

        sequence_features, sentence_features = self._filter_dense_features(
            attribute, featurizers
        )

        combined_sequence_features = self._combine_features(
            sequence_features, featurizers
        )
        combined_sentence_features = self._combine_features(
            sentence_features, featurizers
        )

        return combined_sequence_features, combined_sentence_features

    def get_all_features(
        self, attribute: Text, featurizers: Optional[List[Text]] = None
    ) -> List["Features"]:
        """Gets all features for the attribute given the list of featurizers.

        If no featurizers are provided, all available features will be considered.

        Args:
            attribute: message attribute
            featurizers: names of featurizers to consider

        Returns:
            Features.
        """
        sparse_features = self.get_sparse_features(attribute, featurizers)
        dense_features = self.get_dense_features(attribute, featurizers)

        return [f for f in sparse_features + dense_features if f is not None]

    def features_present(
        self, attribute: Text, featurizers: Optional[List[Text]] = None
    ) -> bool:
        """Checks if there are any features present for the attribute and featurizers.

        If no featurizers are provided, all available features will be considered.

        Args:
            attribute: Message attribute.
            featurizers: Names of featurizers to consider.

        Returns:
            ``True``, if features are present, ``False`` otherwise.
        """
        if featurizers is None:
            featurizers = []

        (
            sequence_sparse_features,
            sentence_sparse_features,
        ) = self._filter_sparse_features(attribute, featurizers)
        sequence_dense_features, sentence_dense_features = self._filter_dense_features(
            attribute, featurizers
        )

        return (
            len(sequence_sparse_features) > 0
            or len(sentence_sparse_features) > 0
            or len(sequence_dense_features) > 0
            or len(sentence_dense_features) > 0
        )

    def _filter_dense_features(
        self, attribute: Text, featurizers: List[Text]
    ) -> Tuple[List["Features"], List["Features"]]:
        sentence_features = [
            f
            for f in self.features
            if f.attribute == attribute
            and f.is_dense()
            and f.type == FEATURE_TYPE_SENTENCE
            and (f.origin in featurizers or not featurizers)
        ]
        sequence_features = [
            f
            for f in self.features
            if f.attribute == attribute
            and f.is_dense()
            and f.type == FEATURE_TYPE_SEQUENCE
            and (f.origin in featurizers or not featurizers)
        ]
        return sequence_features, sentence_features

    def _filter_sparse_features(
        self, attribute: Text, featurizers: List[Text]
    ) -> Tuple[List["Features"], List["Features"]]:
        sentence_features = [
            f
            for f in self.features
            if f.attribute == attribute
            and f.is_sparse()
            and f.type == FEATURE_TYPE_SENTENCE
            and (f.origin in featurizers or not featurizers)
        ]
        sequence_features = [
            f
            for f in self.features
            if f.attribute == attribute
            and f.is_sparse()
            and f.type == FEATURE_TYPE_SEQUENCE
            and (f.origin in featurizers or not featurizers)
        ]

        return sequence_features, sentence_features

    @staticmethod
    def _combine_features(
        features: List["Features"], featurizers: List[Text]
    ) -> Optional["Features"]:
        combined_features = None

        for f in features:
            if combined_features is None:
                combined_features = copy.deepcopy(f)
                combined_features.origin = featurizers
            else:
                combined_features.combine_with_features(f)

        return combined_features

    def is_core_or_domain_message(self) -> bool:
        """Checks whether the message is a core message or from the domain.

        E.g. a core message is created from a story or a domain action,
        not from the NLU data.

        Returns:
            True, if message is a core or domain message, false otherwise.
        """
        return bool(
            self.data.get(ACTION_NAME)
            or self.data.get(ACTION_TEXT)
            or (
                (self.data.get(INTENT) or self.data.get(RESPONSE))
                and not self.data.get(TEXT)
            )
            or (
                self.data.get(TEXT)
                and not (self.data.get(INTENT) or self.data.get(RESPONSE))
            )
        )

    def is_e2e_message(self) -> bool:
        """Checks whether the message came from an e2e story.

        Returns:
            `True`, if message is a from an e2e story, `False` otherwise.
        """
        return bool(
            (self.get(ACTION_TEXT) and not self.get(ACTION_NAME))
            or (self.get(TEXT) and not self.get(INTENT))
        )

    def find_overlapping_entities(
        self,
    ) -> List[Tuple[Dict[Text, Any], Dict[Text, Any]]]:
        """Finds any overlapping entity annotations."""
        entities = self.get(ENTITIES, [])[:]
        entities_with_location = [
            e
            for e in entities
            if (ENTITY_ATTRIBUTE_START in e.keys() and ENTITY_ATTRIBUTE_END in e.keys())
        ]
        entities_with_location.sort(key=lambda e: e[ENTITY_ATTRIBUTE_START])
        overlapping_pairs: List[Tuple[Dict[Text, Any], Dict[Text, Any]]] = []
        for i, entity in enumerate(entities_with_location):
            for other_entity in entities_with_location[i + 1 :]:
                if other_entity[ENTITY_ATTRIBUTE_START] < entity[ENTITY_ATTRIBUTE_END]:
                    overlapping_pairs.append((entity, other_entity))
                else:
                    break
        return overlapping_pairs

    def has_intent(self) -> bool:
        """Checks if the message has an intent."""
        return self.data.get(INTENT) is not None

    def has_commands(self) -> bool:
        """Checks if the message has any commands."""
        return self.data.get(COMMANDS) is not None

    def starts_with_slash_syntax(self) -> bool:
        """Checks if the message text starts with a slash syntax."""
        return self.data.get(TEXT, "").strip().startswith("/")



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Text

if TYPE_CHECKING:
    from rasa.e2e_test.assertions import AssertionFailure
    from rasa.e2e_test.e2e_test_case import TestCase
    from rasa.shared.core.flows.flow_path import FlowPath

NO_RESPONSE = "* No Bot Response *"
NO_SLOT = "* No Slot Set *"


@dataclass
class TestResult:
    """Class for storing results of every test case run."""

    test_case: "TestCase"
    pass_status: bool
    difference: List[Text]
    error_line: Optional[int] = None
    assertion_failure: Optional["AssertionFailure"] = None
    tested_paths: Optional[List["FlowPath"]] = None
    tested_commands: Optional[Dict[str, Dict[str, int]]] = None

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the test result as a dictionary."""
        serialized_test_result = {
            "name": self.test_case.name,
            "pass_status": self.pass_status,
            "expected_steps": [s.as_dict() for s in self.test_case.steps],
        }

        if self.error_line is not None:
            serialized_test_result["error_path"] = (
                f"{self.test_case.file}:{self.error_line}"
            )

        if self.difference:
            serialized_test_result["difference"] = self.difference

        if self.assertion_failure is not None:
            serialized_test_result["assertion_failure"] = (
                self.assertion_failure.as_dict()
            )

        return serialized_test_result


@dataclass
class TestFailure:
    """Class for storing the test case failure."""

    test_case: "TestCase"
    error_line: Optional[int]



from __future__ import annotations

from typing import Any, Dict, Text

from rasa.engine.graph import ExecutionContext, GraphComponent
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.core.flows.yaml_flows_io import YAMLFlowsReader, YamlFlowsWriter

from rasa.shared.core.flows import FlowsList

FLOWS_PERSISTENCE_FILE_NAME = "flows.yml"


class FlowsProvider(GraphComponent):
    """Provides flows information during training and inference time."""

    def __init__(
        self,
        model_storage: ModelStorage,
        resource: Resource,
        flows: FlowsList,
    ) -> None:
        """Creates flows provider."""
        self._model_storage = model_storage
        self._resource = resource
        self._flows = flows

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> FlowsProvider:
        """Creates component (see parent class for full docstring)."""
        return cls(model_storage, resource, flows=FlowsList(underlying_flows=[]))

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> FlowsProvider:
        """Creates provider using a persisted version of itself."""
        with model_storage.read_from(resource) as resource_directory:
            flows = YAMLFlowsReader.read_from_file(
                resource_directory / FLOWS_PERSISTENCE_FILE_NAME
            )
        flows.validate()
        return cls(model_storage, resource, flows)

    def _persist(self, flows: FlowsList) -> None:
        """Persists flows to model storage."""
        with self._model_storage.write_to(self._resource) as resource_directory:
            YamlFlowsWriter.dump(
                flows.underlying_flows,
                resource_directory / FLOWS_PERSISTENCE_FILE_NAME,
            )

    def provide_train(self, importer: TrainingDataImporter) -> FlowsList:
        """Provides flows configuration from training data during training."""
        self._flows = importer.get_flows()
        self._persist(self._flows)
        return self._flows

    def provide_inference(self) -> FlowsList:
        """Provides the flows configuration during inference."""
        return self._flows



from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Text

import dask.local
import dask.core
import asyncio

from rasa.engine.exceptions import GraphRunError
from rasa.engine.graph import ExecutionContext, GraphNode, GraphNodeHook, GraphSchema
from rasa.engine.runner.interface import GraphRunner
from rasa.engine.storage.storage import ModelStorage

logger = logging.getLogger(__name__)


class DaskGraphRunner(GraphRunner):
    """Dask implementation of a `GraphRunner`."""

    def __init__(
        self,
        graph_schema: GraphSchema,
        model_storage: ModelStorage,
        execution_context: ExecutionContext,
        hooks: Optional[List[GraphNodeHook]] = None,
    ) -> None:
        """Initializes a `DaskGraphRunner`.

        Args:
            graph_schema: The graph schema that will be run.
            model_storage: Storage which graph components can use to persist and load
                themselves.
            execution_context: Information about the current graph run to be passed to
                each node.
            hooks: These are called before and after the execution of each node.
        """
        self._graph_schema = graph_schema
        self._instantiated_nodes: Dict[Text, GraphNode] = self._instantiate_nodes(
            graph_schema, model_storage, execution_context, hooks
        )
        self._execution_context: ExecutionContext = execution_context

    @classmethod
    def create(
        cls,
        graph_schema: GraphSchema,
        model_storage: ModelStorage,
        execution_context: ExecutionContext,
        hooks: Optional[List[GraphNodeHook]] = None,
    ) -> DaskGraphRunner:
        """Creates the runner (see parent class for full docstring)."""
        return cls(graph_schema, model_storage, execution_context, hooks)

    @staticmethod
    def _instantiate_nodes(
        graph_schema: GraphSchema,
        model_storage: ModelStorage,
        execution_context: ExecutionContext,
        hooks: Optional[List[GraphNodeHook]] = None,
    ) -> Dict[Text, GraphNode]:
        return {
            node_name: GraphNode.from_schema_node(
                node_name, schema_node, model_storage, execution_context, hooks
            )
            for node_name, schema_node in graph_schema.nodes.items()
        }

    def _build_dask_graph(self, schema: GraphSchema) -> Dict[Text, Any]:
        """Builds a dask graph from the instantiated graph.

        For more information about dask graphs
        see: https://docs.dask.org/en/latest/spec.html
        """
        run_graph = {
            node_name: (
                self._instantiated_nodes[node_name],
                *schema_node.needs.values(),
            )
            for node_name, schema_node in schema.nodes.items()
        }
        return run_graph

    async def run(
        self,
        inputs: Optional[Dict[Text, Any]] = None,
        targets: Optional[List[Text]] = None,
    ) -> Dict[Text, Any]:
        """Runs the graph (see parent class for full docstring)."""
        run_targets = targets if targets else self._graph_schema.target_names
        minimal_schema = self._graph_schema.minimal_graph_schema(run_targets)
        run_graph = self._build_dask_graph(minimal_schema)

        if inputs:
            self._add_inputs_to_graph(inputs, run_graph)

        logger.debug(
            f"Running graph with inputs: {inputs}, targets: {targets} "
            f"and {self._execution_context}."
        )

        try:
            dask_result = await execute_dask_graph(run_graph, run_targets)
            return dict(dask_result)
        except KeyError as e:
            raise GraphRunError(
                f"Could not find key {e} in the graph. Error running runner. "
                f"Please check that you are running bot developed with CALM instead "
                f"of bot developed with previous version of dialog management (DM1)."
            ) from e
        except RuntimeError as e:
            raise GraphRunError("Error running runner.") from e

    @staticmethod
    def _add_inputs_to_graph(inputs: Optional[Dict[Text, Any]], graph: Any) -> None:
        if inputs is None:
            return

        for input_name, input_value in inputs.items():
            if isinstance(input_value, str) and input_value in graph.keys():
                raise GraphRunError(
                    f"Input value '{input_value}' clashes with a node name. Make sure "
                    f"that none of the input names passed to the `run` method are the "
                    f"same as node names in the graph schema."
                )
            graph[input_name] = (input_name, input_value)


async def _execute_task(arg: Any, cache: Dict[str, Any]) -> Any:
    """Do the actual work of collecting data and executing a function.

    Examples:
        >>> inc = lambda x: x + 1
        >>> add = lambda x, y: x + y
        >>> cache = {'x': 1, 'y': 2}

        Compute tasks against a cache
        >>> _execute_task((add, 'x', 1), cache)  # Compute task in naive manner
        2
        >>> _execute_task((add, (inc, 'x'), 1), cache)  # Support nested computation
        3

        Also grab data from cache
        >>> _execute_task('x', cache)
        1

        Support nested lists
        >>> list(_execute_task(['x', 'y'], cache))
        [1, 2]

        >>> list(map(list, _execute_task([['x', 'y'], ['y', 'x']], cache)))
        [[1, 2], [2, 1]]

        >>> _execute_task('foo', cache)  # Passes through on non-keys
        'foo'

    Args:
        arg: The argument to execute (either a function and args or a value)
        cache: A cache to store intermediate results

    Returns:
        The result of the computation.
    """
    if isinstance(arg, list):
        return [await _execute_task(a, cache) for a in arg]
    elif dask.core.istask(arg):  # type:ignore[no-untyped-call]
        func, args = arg[0], arg[1:]
        # Note: Don't assign the subtask results to a variable. numpy detects
        # temporaries by their reference count and can execute certain
        # operations in-place.
        awaited_args = await asyncio.gather(*(_execute_task(a, cache) for a in args))
        if hasattr(func, "__call__") and asyncio.iscoroutinefunction(func.__call__):
            # in most cases, `func` will be an instance of `GraphNode`. rather
            # than a function directly. `GraphNode` instances have a `__call__`
            # method that is a coroutine function and will be called here.
            return await func(*awaited_args)
        else:
            # This is based on the original dask implementation.
            # I (tmbo) do not think that we will ever go down that else path,
            # but I kept it for compatibility. All our graph nodes should be
            # of type `GraphNode` and end up in the above if. But honestly,
            # I am not 100% sure they always are, so I wanted to make sure
            # that if there is a node that is not a GraphNode, it runs
            # just fine with the prior behaviour.
            return func(*awaited_args)
    elif not dask.core.ishashable(arg):  # type:ignore[no-untyped-call]
        return arg
    elif arg in cache:
        return cache[arg]
    else:
        return arg


async def execute_dask_graph(dsk: Dict[str, Any], result: List[str]) -> Any:
    """Asynchronous get function.

    This is a general version of various asynchronous schedulers for dask.  It
    takes a ``concurrent.futures.Executor.submit`` function to form a more
    specific ``get`` method that walks through the dask array with parallel
    workers, avoiding repeat computation and minimizing memory use.

    Build on the version of `get_async` in `dask.local`. This version is
    asynchronous and uses `asyncio` to manage the event loop. It is designed to
    be used in an `async` function.

    The default dask implementation uses an internal event loop, which means
    blocking the outer event loop (e.g. sanic or webhooks).

    Args:
        dsk: A dask dictionary specifying a workflow
        result: Keys corresponding to desired data

    Returns:
        The result keys values after computing the graph.
    """
    cache = None

    results = set(result)

    # if start_state_from_dask fails, we will have something
    # to pass to the final block.
    state = {}
    keyorder = dask.local.order(dsk)  # type:ignore[no-untyped-call]

    state = dask.local.start_state_from_dask(dsk, cache=cache, sortkey=keyorder.get)  # type:ignore[no-untyped-call]

    if state["waiting"] and not state["ready"]:
        raise ValueError("Found no accessible jobs in dask")

    async def fire_task() -> None:
        """Fire off a task to the thread pool."""
        # start a new job
        # Get the next task to compute (most recently added)
        key = state["ready"].pop()
        # Notify task is running
        state["running"].add(key)

        dependencies = dask.local.get_dependencies(dsk, key)  # type:ignore[no-untyped-call]
        # Prep args to send
        data = {dep: state["cache"][dep] for dep in dependencies}

        task_result = await _execute_task(dsk[key], data)
        state["cache"][key] = task_result
        dask.local.finish_task(dsk, key, state, results, keyorder.get)  # type:ignore[no-untyped-call]

    # Main loop, wait on tasks to finish, insert new ones
    while state["ready"]:
        await fire_task()

    return dask.local.nested_get(result, state["cache"])  # type:ignore[no-untyped-call]



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Text, Type

import rasa.shared.core.constants
from rasa.shared.exceptions import RasaException
import rasa.shared.utils.common
import rasa.shared.utils.io
from rasa.shared.constants import (
    DOCS_URL_SLOTS,
    DOCS_URL_NLU_BASED_SLOTS,
    DOCS_URL_CATEGORICAL_SLOTS,
)

logger = logging.getLogger(__name__)


class InvalidSlotTypeException(RasaException):
    """Raised if a slot type is invalid."""


class InvalidSlotConfigError(RasaException, ValueError):
    """Raised if a slot's config is invalid."""


class Slot(ABC):
    """Key-value store for storing information during a conversation."""

    @property
    @abstractmethod
    def type_name(self) -> Text:
        """Name of the type of slot."""
        ...

    def __init__(
        self,
        name: Text,
        mappings: List[Dict[Text, Any]],
        initial_value: Any = None,
        value_reset_delay: Optional[int] = None,
        influence_conversation: bool = True,
        is_builtin: bool = False,
        shared_for_coexistence: bool = False,
    ) -> None:
        """Create a Slot.

        Args:
            name: The name of the slot.
            initial_value: The initial value of the slot.
            mappings: List containing slot mappings.
            value_reset_delay: After how many turns the slot should be reset to the
                initial_value. This is behavior is currently not implemented.
            influence_conversation: If `True` the slot will be featurized and hence
                influence the predictions of the dialogue polices.
            is_builtin: `True` if the slot is a built-in slot, `False` otherwise.
                Built-in slots also encompass user writable slots (via custom actions),
                such as `return_value`.
            shared_for_coexistence: If `True` the slot is not forgotten after either
                dm1 or CALM finishes.
        """
        self.name = name
        self.mappings = mappings
        self._value = initial_value
        self.initial_value = initial_value
        self._value_reset_delay = value_reset_delay
        self.influence_conversation = influence_conversation
        self._has_been_set = False
        self.is_builtin = is_builtin
        self.shared_for_coexistence = shared_for_coexistence

    def feature_dimensionality(self) -> int:
        """How many features this single slot creates.

        Returns:
            The number of features. `0` if the slot is unfeaturized. The dimensionality
            of the array returned by `as_feature` needs to correspond to this value.
        """
        if not self.influence_conversation:
            return 0

        return self._feature_dimensionality()

    def _feature_dimensionality(self) -> int:
        """See the docstring for `feature_dimensionality`."""
        return 1

    def has_features(self) -> bool:
        """Indicate if the slot creates any features."""
        return self.feature_dimensionality() != 0

    def value_reset_delay(self) -> Optional[int]:
        """After how many turns the slot should be reset to the initial_value.

        If the delay is set to `None`, the slot will keep its value forever.
        """
        # TODO: FUTURE this needs to be implemented - slots are not reset yet
        return self._value_reset_delay

    def as_feature(self) -> List[float]:
        if not self.influence_conversation:
            return []

        return self._as_feature()

    @abstractmethod
    def _as_feature(self) -> List[float]:
        raise NotImplementedError(
            "Each slot type needs to specify how its "
            "value can be converted to a feature. Slot "
            "'{}' is a generic slot that can not be used "
            "for predictions. Make sure you add this "
            "slot to your domain definition, specifying "
            "the type of the slot. If you implemented "
            "a custom slot type class, make sure to "
            "implement `.as_feature()`."
            "".format(self.name)
        )

    def reset(self) -> None:
        """Resets the slot's value to the initial value."""
        self.value = self.initial_value
        self._has_been_set = False

    @property
    def value(self) -> Any:
        """Gets the slot's value."""
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        """Sets the slot's value."""
        self._value = value
        self._has_been_set = True

    def has_same_coerced_value(self, other_value: Any) -> bool:
        """Checks if the coerced value of is the same as the slot value.

        Can be used to check if an update with the value would be a no-op.
        """
        return self.coerce_value(other_value) == self.value

    def coerce_value(self, value: Any) -> Any:
        """Coerces the value to the slot's type."""
        return value

    def is_valid_value(self, value: Any) -> bool:
        """Checks if the slot contains the value."""
        return True

    @property
    def has_been_set(self) -> bool:
        """Indicates if the slot's value has been set."""
        return self._has_been_set

    def __str__(self) -> Text:
        return f"{self.__class__.__name__}({self.name}: {self.value})"

    def __repr__(self) -> Text:
        return f"<{self.__class__.__name__}({self.name}: {self.value})>"

    @staticmethod
    def resolve_by_type(type_name: Text) -> Type["Slot"]:
        """Returns a slots class by its type name."""
        for cls in rasa.shared.utils.common.all_subclasses(Slot):
            if cls.type_name == type_name:
                return cls
        try:
            return rasa.shared.utils.common.class_from_module_path(type_name)
        except (ImportError, AttributeError):
            raise InvalidSlotTypeException(
                f"Failed to find slot type, '{type_name}' is neither a known type nor "
                f"user-defined. If you are creating your own slot type, make "
                f"sure its module path is correct. "
                f"You can find all build in types at {DOCS_URL_SLOTS}"
            )

    def persistence_info(self) -> Dict[str, Any]:
        """Returns relevant information to persist this slot."""
        return {
            "type": rasa.shared.utils.common.module_path_from_instance(self),
            "initial_value": self.initial_value,
            "influence_conversation": self.influence_conversation,
            "mappings": self.mappings,
        }

    def fingerprint(self) -> Text:
        """Returns a unique hash for the slot which is stable across python runs.

        Returns:
            fingerprint of the slot
        """
        data = {"slot_name": self.name, "slot_value": self.value}
        data.update(self.persistence_info())
        return rasa.shared.utils.io.get_dictionary_fingerprint(data)


class FloatSlot(Slot):
    """A slot storing a float value."""

    type_name = "float"

    def __init__(
        self,
        name: Text,
        mappings: List[Dict[Text, Any]],
        initial_value: Optional[float] = None,
        value_reset_delay: Optional[int] = None,
        max_value: float = 1.0,
        min_value: float = 0.0,
        influence_conversation: bool = True,
        is_builtin: bool = False,
        shared_for_coexistence: bool = False,
    ) -> None:
        """Creates a FloatSlot.

        Raises:
            InvalidSlotConfigError, if the min-max range is invalid.
            UserWarning, if initial_value is outside the min-max range.
        """
        super().__init__(
            name,
            mappings,
            initial_value,
            value_reset_delay,
            influence_conversation,
            is_builtin,
            shared_for_coexistence,
        )
        self.max_value = max_value
        self.min_value = min_value

        if min_value >= max_value:
            raise InvalidSlotConfigError(
                "Float slot ('{}') created with an invalid range "
                "using min ({}) and max ({}) values. Make sure "
                "min is smaller than max."
                "".format(self.name, self.min_value, self.max_value)
            )

        if initial_value is not None and not (min_value <= initial_value <= max_value):
            rasa.shared.utils.io.raise_warning(
                f"Float slot ('{self.name}') created with an initial value "
                f"{self.value}. This value is outside of the configured min "
                f"({self.min_value}) and max ({self.max_value}) values."
            )

    def _as_feature(self) -> List[float]:
        try:
            capped_value = max(self.min_value, min(self.max_value, float(self.value)))
            if abs(self.max_value - self.min_value) > 0:
                covered_range = abs(self.max_value - self.min_value)
            else:
                covered_range = 1
            return [1.0, (capped_value - self.min_value) / covered_range]
        except (TypeError, ValueError):
            return [0.0, 0.0]

    def persistence_info(self) -> Dict[Text, Any]:
        """Returns relevant information to persist this slot."""
        d = super().persistence_info()
        d["max_value"] = self.max_value
        d["min_value"] = self.min_value
        return d

    def coerce_value(self, value: Any) -> Any:
        """Coerces the value to the slot's type."""
        try:
            return float(value)
        except (TypeError, ValueError):
            return value

    def is_valid_value(self, value: Any) -> bool:
        """Checks if the slot contains the value."""
        # check that coerced type is float
        return value is None or isinstance(self.coerce_value(value), float)

    def _feature_dimensionality(self) -> int:
        return len(self.as_feature())


class BooleanSlot(Slot):
    """A slot storing a truth value."""

    type_name = "bool"

    def _as_feature(self) -> List[float]:
        try:
            if self.value is not None:
                return [1.0, float(bool_from_any(self.value))]
            else:
                return [0.0, 0.0]
        except (TypeError, ValueError):
            # we couldn't convert the value to float - using default value
            return [0.0, 0.0]

    def _feature_dimensionality(self) -> int:
        return len(self.as_feature())

    def coerce_value(self, value: Any) -> Any:
        """Coerces the value to the slot's type."""
        try:
            return bool_from_any(value)
        except (ValueError, TypeError):
            return value

    def is_valid_value(self, value: Any) -> bool:
        """Checks if the slot contains the value."""
        return value is None or isinstance(self.coerce_value(value), bool)


def bool_from_any(x: Any) -> bool:
    """Converts bool/float/int/str to bool or raises error."""
    if isinstance(x, bool):
        return x
    elif isinstance(x, (float, int)):
        return x == 1.0
    elif isinstance(x, str):
        if x.isnumeric():
            return float(x) == 1.0
        elif x.strip().lower() == "true":
            return True
        elif x.strip().lower() == "no":
            return False
        elif x.strip().lower() == "false":
            return False
        elif x.strip().lower() == "yes":
            return True
        else:
            raise ValueError("Cannot convert string to bool")
    else:
        raise TypeError("Cannot convert to bool")


class TextSlot(Slot):
    type_name = "text"

    def _as_feature(self) -> List[float]:
        return [1.0 if self.value is not None else 0.0]


class ListSlot(Slot):
    type_name = "list"

    def _as_feature(self) -> List[float]:
        try:
            if self.value is not None and len(self.value) > 0:
                return [1.0]
            else:
                return [0.0]
        except (TypeError, ValueError):
            # we couldn't convert the value to a list - using default value
            return [0.0]

    def coerce_value(self, value: Any) -> Any:
        """Coerces the value to the slot's type."""
        if value and not isinstance(value, list):
            # Make sure we always store list items
            value = [value]
        return value

    # FIXME: https://github.com/python/mypy/issues/8085
    @Slot.value.setter  # type: ignore[attr-defined,misc]
    def value(self, value: Any) -> None:
        """Sets the slot's value."""
        # Call property setter of superclass
        # FIXME: https://github.com/python/mypy/issues/8085
        super(ListSlot, self.__class__).value.fset(self, self.coerce_value(value))  # type: ignore[attr-defined]


class CategoricalSlot(Slot):
    """Slot type which can be used to branch conversations based on its value."""

    type_name = "categorical"

    def __init__(
        self,
        name: Text,
        mappings: List[Dict[Text, Any]],
        values: Optional[List[Any]] = None,
        initial_value: Any = None,
        value_reset_delay: Optional[int] = None,
        influence_conversation: bool = True,
        is_builtin: bool = False,
        shared_for_coexistence: bool = False,
    ) -> None:
        """Creates a `Categorical  Slot` (see parent class for detailed docstring)."""
        super().__init__(
            name,
            mappings,
            initial_value,
            value_reset_delay,
            influence_conversation,
            is_builtin,
            shared_for_coexistence,
        )
        if values and None in values:
            rasa.shared.utils.io.raise_warning(
                f"Categorical slot '{self.name}' has `null` listed as a possible value"
                f" in the domain file, which translates to `None` in Python. This value"
                f" is reserved for when the slot is not set, and should not be listed"
                f" as a value in the slot's definition."
                f" Rasa will ignore `null` as a possible value for the '{self.name}'"
                f" slot. Consider changing this value in your domain file to, for"
                f" example, `unset`, or provide the value explicitly as a string by"
                f' using quotation marks: "null".',
                category=UserWarning,
            )
        self.values = [str(v) for v in values if v is not None] if values else []
        self._warn_for_duplicate_coerced_values(values)

    def _warn_for_duplicate_coerced_values(
        self, possible_values: Optional[List[Any]]
    ) -> None:
        """If multiple values are coerced to the same value, warn the user.

        A warning indicates that multiple values are coerced to the same value
        which can create unexpected behavior. For example, if the slot is
        configured with values `["a", "A"]` both of these values are coerced
        to `"a"`. This can lead to unexpected behavior if the user sets the
        slot to `"A"` and the slot is later checked for the value `"a"`.

        Args:
            possible_values: The values for this slot configured in the domain.

        Returns:
            None
        """
        if not possible_values:
            return

        coerced_values: Dict[str, Any] = {}
        for value in possible_values:
            if value is None:
                continue

            if isinstance(value, Text):
                coerced_value = value.casefold()
            else:
                # Make sure we always store items as strings - though these
                # could be dicts, so we do not want to lower them
                coerced_value = str(value)

            if coerced_value in coerced_values:
                rasa.shared.utils.io.raise_warning(
                    f"Multiple values are coerced to the same value for the "
                    f"categorical slot '{self.name}'. "
                    f"Coerced value '{coerced_value}' is the same for "
                    f"values '{value}' and '{coerced_values[coerced_value]}'. "
                    f"Make sure to provide unique values for the slot.",
                    category=UserWarning,
                    docs=DOCS_URL_CATEGORICAL_SLOTS,
                )
            coerced_values[coerced_value] = value

    def add_default_value(self) -> None:
        """Adds the special default value to the list of possible values."""
        values = set(self.values)
        if rasa.shared.core.constants.DEFAULT_CATEGORICAL_SLOT_VALUE not in values:
            self.values.append(
                rasa.shared.core.constants.DEFAULT_CATEGORICAL_SLOT_VALUE
            )

    def persistence_info(self) -> Dict[Text, Any]:
        """Returns serialized slot."""
        d = super().persistence_info()
        d["values"] = [
            value
            for value in self.values
            # Don't add default slot when persisting it.
            # We'll re-add it on the fly when creating the domain.
            if value != rasa.shared.core.constants.DEFAULT_CATEGORICAL_SLOT_VALUE
        ]
        return d

    def _as_feature(self) -> List[float]:
        r = [0.0] * self.feature_dimensionality()

        # Return the zero-filled array if the slot is unset (i.e. set to None).
        # Conceptually, this is similar to the case when the featurisation process
        # fails, hence the returned features here are the same as for that case.
        if self.value is None:
            return r

        try:
            for i, v in enumerate(self.values):
                if v.lower() == str(self.value).lower():
                    r[i] = 1.0
                    break
            else:
                if (
                    rasa.shared.core.constants.DEFAULT_CATEGORICAL_SLOT_VALUE
                    in self.values
                ):
                    i = self.values.index(
                        rasa.shared.core.constants.DEFAULT_CATEGORICAL_SLOT_VALUE
                    )
                    r[i] = 1.0
                else:
                    rasa.shared.utils.io.raise_warning(
                        f"Categorical slot '{self.name}' is set to a value "
                        f"('{self.value}') "
                        "that is not specified in the domain. "
                        "Value will be ignored and the slot will "
                        "behave as if no value is set. "
                        "Make sure to add all values a categorical "
                        "slot should store to the domain."
                    )
        except (TypeError, ValueError):
            logger.exception("Failed to featurize categorical slot.")
            return r
        return r

    def _feature_dimensionality(self) -> int:
        return len(self.values)

    def coerce_value(self, value: Any) -> Any:
        """Coerces the value to the slot's configured values if possible.

        If the value can not be mapped to a configured value, the value
        is returned as is.

        String values are compared case insensitive. For example, if the
        slot is configured with values `["a", "b"]` and the value is `"A"`,
        the value is coerced to `"a"`. One exception is the case where
        the exact value is configured in the domain. In this case, the value's
        casing is not changed. For example, if the slot is configured with
        values `["A", "B"]` and the value is `"A"`, the value is coerced to
        `"A"`.

        Args:
            value: The value to coerce.

        Returns:
            The coerced value or the value as is.
        """
        if value and isinstance(value, Text):
            # Make sure we always store items with the casing
            # as they are defined in the domain
            # first, try to find an exact match
            if value in self.values:
                return value

            # fallback: try to find a case insensitive match
            for configured_value in self.values:
                if value.casefold() == configured_value.casefold():
                    return configured_value
        return value

    def is_valid_value(self, value: Any) -> bool:
        """Checks if the slots configured values contains the value.

        If the value is `None`, it is always considered valid.

        Args:
            value: The value to check.

        Returns:
            `True` if the value is valid, `False` otherwise.
        """
        if value is None:
            return True
        if isinstance(value, Text):
            # Make sure we always store items with the casing
            # as they are defined in the domain
            for configured_value in self.values:
                if value.casefold() == configured_value.casefold():
                    return True
            return False

        return str(value) in self.values

    # FIXME: https://github.com/python/mypy/issues/8085
    @Slot.value.setter  # type: ignore[attr-defined,misc]
    def value(self, value: Any) -> None:
        """Sets the slot's value after coercing it to a configured value.

        If the value can not be mapped to a configured value, the value
        is kept as is.
        """
        # Call property setter of superclass
        # FIXME: https://github.com/python/mypy/issues/8085
        super(CategoricalSlot, self.__class__).value.fset(  # type: ignore[attr-defined]
            self, self.coerce_value(value)
        )


class AnySlot(Slot):
    """Slot which can be used to store any value.

    Users need to create a subclass of `Slot` in case
    the information is supposed to get featurized.
    """

    type_name = "any"

    def __init__(
        self,
        name: Text,
        mappings: List[Dict[Text, Any]],
        initial_value: Any = None,
        value_reset_delay: Optional[int] = None,
        influence_conversation: bool = False,
        is_builtin: bool = False,
        shared_for_coexistence: bool = False,
    ) -> None:
        """Creates an `Any  Slot` (see parent class for detailed docstring).

        Raises:
            InvalidSlotConfigError, if slot is featurized.
        """
        if influence_conversation:
            raise InvalidSlotConfigError(
                f"An {AnySlot.__name__} cannot be featurized. "
                f"Please use a different slot type for slot '{name}' instead. If you "
                f"need to featurize a data type which is not supported out of the box, "
                f"implement a custom slot type by subclassing '{Slot.__name__}'. See "
                f"the documentation for more information: {DOCS_URL_NLU_BASED_SLOTS}"
            )

        super().__init__(
            name,
            mappings,
            initial_value,
            value_reset_delay,
            influence_conversation,
            is_builtin,
            shared_for_coexistence,
        )

    def __eq__(self, other: Any) -> bool:
        """Compares object with other object."""
        if not isinstance(other, AnySlot):
            return NotImplemented

        return (
            self.name == other.name
            and self.initial_value == other.initial_value
            and self._value_reset_delay == other._value_reset_delay
            and self.value == other.value
        )

    def _as_feature(self) -> List[float]:
        raise InvalidSlotConfigError(
            f"An {AnySlot.__name__} cannot be featurized. "
            f"Please use a different slot type for slot '{self.name}' instead. If you "
            f"need to featurize a data type which is not supported out of the box, "
            f"implement a custom slot type by subclassing '{Slot.__name__}'. "
            f"See the documentation for more information: {DOCS_URL_NLU_BASED_SLOTS}"
        )



from __future__ import annotations

import datetime
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Text, Union

import jwt
from keycloak import KeycloakOpenID, KeycloakError
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.yaml import read_yaml_file, write_yaml

from rasa.studio.config import StudioConfig
from rasa.studio.constants import (
    KEYCLOAK_ACCESS_TOKEN_KEY,
    KEYCLOAK_EXPIRES_IN_KEY,
    KEYCLOAK_REFRESH_EXPIRES_IN_KEY,
    KEYCLOAK_REFRESH_TOKEN,
)
from rasa.studio.results_logger import with_studio_error_handler, StudioResult


class StudioAuth:
    """Handles the authentication with the Rasa Studio authentication server."""

    def __init__(self, studio_config: StudioConfig) -> None:
        self.config = studio_config
        verify = not studio_config.disable_verify

        self.keycloak_openid = KeycloakOpenID(
            server_url=studio_config.authentication_server_url,
            client_id=studio_config.client_id,
            realm_name=studio_config.realm_name,
            verify=verify,
        )

    def health_check(self) -> bool:
        """Check if the Keycloak server is reachable.

        Returns:
        True if the server is reachable, False otherwise.
        """
        try:
            self.keycloak_openid.well_known()
            return True
        except Exception:
            return False

    @with_studio_error_handler
    def login(
        self, username: Text, password: Text, totp: Optional[int] = None
    ) -> StudioResult:
        token_dict = self.keycloak_openid.token(
            username=username, password=password, totp=totp
        )
        keycloak_token = self._resolve_token(token_dict)

        KeycloakTokenWriter.write_token_to_file(
            keycloak_token, token_file_location=DEFAULT_TOKEN_FILE_PATH
        )

        return StudioResult.success("Login successful.")

    @with_studio_error_handler
    def refresh_token(self, refresh_token: Text) -> StudioResult:
        try:
            token_dict = self.keycloak_openid.refresh_token(refresh_token)
        except Exception as e:
            raise KeycloakError(f"Could not refresh token. Error: {e}")

        keycloak_token = self._resolve_token(token_dict)

        KeycloakTokenWriter.write_token_to_file(
            keycloak_token, token_file_location=DEFAULT_TOKEN_FILE_PATH
        )

        return StudioResult.success("Token refreshed successfully.")

    @staticmethod
    def _resolve_token(token_dict: Dict[Text, Any]) -> KeycloakToken:
        return KeycloakToken(
            access_token=token_dict[KEYCLOAK_ACCESS_TOKEN_KEY],
            expires_in=token_dict[KEYCLOAK_EXPIRES_IN_KEY],
            refresh_expires_in=token_dict[KEYCLOAK_REFRESH_EXPIRES_IN_KEY],
            refresh_token=token_dict[KEYCLOAK_REFRESH_TOKEN],
            token_type=token_dict[TOKEN_TYPE_KEY],
        )


ACCESS_TOKEN_KEY = "access_token"
ACCESS_TOKEN_EXPIRATION_TIME_KEY = "access_token_expiration_time"
REFRESH_TOKEN_KEY = "refresh_token"
REFRESH_TOKEN_EXPIRATION_TIME_KEY = "refresh_token_expiration_time"
TOKEN_TYPE_KEY = "token_type"

JWT_TOKEN_EXPIRES_AT_KEY = "exp"


@dataclass
class KeycloakToken:
    access_token: Text
    expires_in: int
    refresh_expires_in: int
    refresh_token: Text
    token_type: Text

    def to_dict(self) -> Dict[Text, Union[Text, int]]:
        return {
            ACCESS_TOKEN_KEY: self.access_token,
            ACCESS_TOKEN_EXPIRATION_TIME_KEY: self.expires_in,
            REFRESH_TOKEN_KEY: self.refresh_token,
            REFRESH_TOKEN_EXPIRATION_TIME_KEY: self.refresh_expires_in,
            TOKEN_TYPE_KEY: self.token_type,
        }

    def expires_at(self) -> datetime.datetime:
        jwt_access_token = KeycloakToken._decode_token(self.access_token)
        expires_at = jwt_access_token[JWT_TOKEN_EXPIRES_AT_KEY]
        return datetime.datetime.utcfromtimestamp(expires_at)

    def is_expired(self) -> bool:
        return self.expires_at() < datetime.datetime.now()

    def can_refresh(self) -> bool:
        return self._has_refresh_token() and not self._has_refresh_token_expired()

    def _refresh_expiration_time(self) -> datetime.datetime:
        jwt_refresh_token = KeycloakToken._decode_token(self.refresh_token)
        expires_at = jwt_refresh_token[JWT_TOKEN_EXPIRES_AT_KEY]
        return datetime.datetime.utcfromtimestamp(expires_at)

    def _has_refresh_token_expired(self) -> bool:
        return self._refresh_expiration_time() < datetime.datetime.now()

    def _has_refresh_token(self) -> bool:
        return self.refresh_token is not None and self.refresh_token != ""

    @staticmethod
    def _decode_token(token: Text) -> Dict[Text, Any]:
        """Decode a JWT token."""
        try:
            return jwt.decode(token, options={"verify_signature": False})
        except jwt.exceptions.DecodeError as e:
            raise RasaException(f"Could not decode token. Error: {e}.")


YAMLContent = Union[List[Any], Dict[Text, Any]]

DEFAULT_RASA_CONFIG_PATH = Path.home() / ".config" / "rasa"
DEFAULT_TOKEN_FILE = "studio_token.yaml"
DEFAULT_TOKEN_FILE_PATH = DEFAULT_RASA_CONFIG_PATH / DEFAULT_TOKEN_FILE


class KeycloakTokenReader:
    """Reads the token from the token file.

    If the token file does not exist, exception will be raised.
    """

    def __init__(self, token_path: Path = DEFAULT_TOKEN_FILE_PATH) -> None:
        """Initialise the token reader.

        Args:
            token_path: path to the token file.
        """
        token_file_content = self._load_token_from_cache_file(token_path)
        self.token = self._valid_token_content(token_file_content)

    def get_token(self) -> KeycloakToken:
        """Returns the token from the token file.

        Returns:
            The token from the token file.
        """
        return self.token

    @staticmethod
    def _valid_token_content(content: YAMLContent) -> KeycloakToken:
        if isinstance(content, list):
            raise ValueError(
                "Content of token file is in list format. "
                "Content should be in dictionary format. "
                "Please make sure the file is a valid YAML file."
            )

        access_token = content.get(ACCESS_TOKEN_KEY)
        if not access_token or not isinstance(access_token, str):
            raise ValueError(
                f"{ACCESS_TOKEN_KEY} in token file is not in text format. "
                "Please make sure the value is in text format."
            )

        access_token_expiration_time = content.get(ACCESS_TOKEN_EXPIRATION_TIME_KEY)
        if not isinstance(access_token_expiration_time, int):
            raise ValueError(
                f"{ACCESS_TOKEN_EXPIRATION_TIME_KEY} in "
                f"token file is not in integer format. "
                "Please make sure the value is in integer format."
            )

        refresh_token = content.get(REFRESH_TOKEN_KEY)
        if refresh_token is None:
            raise ValueError(
                f"{REFRESH_TOKEN_KEY} in token file is "
                f"not present or it is set to None. "
                "Please make sure the value is present and in text format."
            )
        elif not isinstance(refresh_token, str) or refresh_token == "":
            raise ValueError(
                f"{REFRESH_TOKEN_KEY} in token file is not in text format. "
                "Please make sure the value is in text format."
            )

        refresh_token_expiration_time = content.get(REFRESH_TOKEN_EXPIRATION_TIME_KEY)
        if refresh_token_expiration_time is None:
            raise ValueError(
                f"{REFRESH_TOKEN_EXPIRATION_TIME_KEY} in token file "
                f"is not present or it is set to None. "
                "Please make sure the value is present and in text format."
            )
        elif (
            not isinstance(refresh_token_expiration_time, int)
            or refresh_token_expiration_time == ""
        ):
            raise ValueError(
                f"{REFRESH_TOKEN_EXPIRATION_TIME_KEY} in "
                f"token file is not in text format. "
                "Please make sure the value is in text format."
            )

        token_type = content.get(TOKEN_TYPE_KEY)

        if token_type is None:
            raise ValueError(
                f"{TOKEN_TYPE_KEY} in token file is "
                f"not present or it is set to None. "
                "Please make sure the value is present and in text format."
            )
        elif not isinstance(token_type, str) or token_type == "":
            raise ValueError(
                f"{TOKEN_TYPE_KEY} in "
                f"token file is not in text format. "
                "Please make sure the value is in text format."
            )

        return KeycloakToken(
            access_token=access_token,
            expires_in=access_token_expiration_time,
            refresh_token=refresh_token,
            refresh_expires_in=refresh_token_expiration_time,
            token_type=token_type,
        )

    @staticmethod
    def _load_token_from_cache_file(
        token_path: Path,
    ) -> YAMLContent:
        if not token_path.is_file():
            raise RasaException(
                f"Could not load Keycloak token from location {token_path}."
                f"Please make sure to run `rasa studio login` first."
            )

        return read_yaml_file(token_path)


class KeycloakTokenWriter:
    @staticmethod
    def write_token_to_file(token: KeycloakToken, token_file_location: Path) -> None:
        write_yaml(token.to_dict(), token_file_location)



import logging
from typing import List, Optional, Dict, Text, Set, Any

import numpy as np
import scipy.sparse

from rasa.core.featurizers.precomputation import MessageContainerForCoreFeaturization
from rasa.nlu.extractors.extractor import EntityTagSpec
from rasa.nlu.utils import bilou_utils
from rasa.nlu.utils.bilou_utils import BILOU_PREFIXES
from rasa.shared.core.domain import SubState, State, Domain
from rasa.shared.core.constants import PREVIOUS_ACTION, ACTIVE_LOOP, USER, SLOTS
from rasa.shared.core.trackers import is_prev_action_listen_in_state
from rasa.shared.nlu.constants import (
    ENTITIES,
    FEATURE_TYPE_SENTENCE,
    ACTION_TEXT,
    ACTION_NAME,
    INTENT,
    NO_ENTITY_TAG,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_TAGS,
    TEXT,
)
from rasa.shared.nlu.training_data.features import Features
from rasa.utils.tensorflow import model_data_utils

logger = logging.getLogger(__name__)


class SingleStateFeaturizer:
    """Base class to transform the dialogue state into an ML format.

    Subclasses of SingleStateFeaturizer will decide how a bot will
    transform the dialogue state into a dictionary mapping an attribute
    to its features. Possible attributes are: `INTENT`, `TEXT`, `ACTION_NAME`,
    `ACTION_TEXT`, `ENTITIES`, `SLOTS` and `ACTIVE_LOOP`. Each attribute will be
    featurized into a list of `rasa.utils.features.Features`.
    """

    def __init__(self) -> None:
        """Initialize the single state featurizer."""
        self._default_feature_states: Dict[Text, Any] = {}
        self.action_texts: List[Text] = []
        self.entity_tag_specs: List[EntityTagSpec] = []

    def _create_entity_tag_specs(
        self, bilou_tagging: bool = False
    ) -> List[EntityTagSpec]:
        """Returns the tag to index mapping for entities.

        Returns:
            Tag to index mapping.
        """
        if ENTITIES not in self._default_feature_states:
            return []

        if bilou_tagging:
            tag_id_index_mapping = {
                f"{prefix}{tag}": idx_1 * len(BILOU_PREFIXES) + idx_2 + 1
                for tag, idx_1 in self._default_feature_states[ENTITIES].items()
                for idx_2, prefix in enumerate(BILOU_PREFIXES)
            }
        else:
            tag_id_index_mapping = {
                tag: idx + 1  # +1 to keep 0 for the NO_ENTITY_TAG
                for tag, idx in self._default_feature_states[ENTITIES].items()
            }

        # NO_ENTITY_TAG corresponds to non-entity which should correspond to 0 index
        # needed for correct prediction for padding
        tag_id_index_mapping[NO_ENTITY_TAG] = 0

        # TODO
        #  The entity states used to create the tag-idx-mapping contains the
        #  entities and the concatenated entity and roles/groups. We do not
        #  distinguish between entities and roles/groups right now.
        #  we return a list to anticipate that
        return [
            EntityTagSpec(
                tag_name=ENTITY_ATTRIBUTE_TYPE,
                tags_to_ids=tag_id_index_mapping,
                ids_to_tags={value: key for key, value in tag_id_index_mapping.items()},
                num_tags=len(tag_id_index_mapping),
            )
        ]

    def prepare_for_training(self, domain: Domain, bilou_tagging: bool = False) -> None:
        """Gets necessary information for featurization from domain.

        Args:
            domain: An instance of :class:`rasa.shared.core.domain.Domain`.
            bilou_tagging: indicates whether BILOU tagging should be used or not
        """

        # store feature states for each attribute in order to create binary features
        def convert_to_dict(feature_states: List[Text]) -> Dict[Text, int]:
            return {
                feature_state: idx for idx, feature_state in enumerate(feature_states)
            }

        self._default_feature_states[INTENT] = convert_to_dict(domain.intents)
        self._default_feature_states[ACTION_NAME] = convert_to_dict(
            domain.action_names_or_texts
        )
        self._default_feature_states[ENTITIES] = convert_to_dict(domain.entity_states)
        self._default_feature_states[SLOTS] = convert_to_dict(domain.slot_states)
        self._default_feature_states[ACTIVE_LOOP] = convert_to_dict(domain.form_names)
        self.action_texts = domain.action_texts
        self.entity_tag_specs = self._create_entity_tag_specs(bilou_tagging)

    def _state_features_for_attribute(
        self, sub_state: SubState, attribute: Text
    ) -> Dict[Text, int]:
        # FIXME: the code below is not type-safe, but fixing it
        #        would require more refactoring, for instance using
        #        data classes in our states
        if attribute in {INTENT, ACTION_NAME}:
            return {sub_state[attribute]: 1}  # type: ignore[dict-item]
        elif attribute == ENTITIES:
            return {entity: 1 for entity in sub_state.get(ENTITIES, [])}  # type: ignore[misc]
        elif attribute == ACTIVE_LOOP:
            return {sub_state["name"]: 1}  # type: ignore[dict-item]
        elif attribute == SLOTS:
            return {
                f"{slot_name}_{i}": value  # type: ignore[misc]
                for slot_name, slot_as_feature in sub_state.items()
                for i, value in enumerate(slot_as_feature)
            }
        else:
            raise ValueError(
                f"Given attribute '{attribute}' is not supported. "
                f"It must be one of '{self._default_feature_states.keys()}'."
            )

    def _create_features(
        self, sub_state: SubState, attribute: Text, sparse: bool = False
    ) -> List[Features]:
        state_features = self._state_features_for_attribute(sub_state, attribute)

        features = np.zeros(len(self._default_feature_states[attribute]), np.float32)
        for state_feature, value in state_features.items():
            # check that the value is in default_feature_states to be able to assign
            # its value
            if state_feature in self._default_feature_states[attribute]:
                features[self._default_feature_states[attribute][state_feature]] = value
        features = np.expand_dims(features, 0)

        if sparse:
            features = scipy.sparse.coo_matrix(features)

        return [
            Features(
                features, FEATURE_TYPE_SENTENCE, attribute, self.__class__.__name__
            )
        ]

    @staticmethod
    def _to_sparse_sentence_features(
        sparse_sequence_features: List[Features],
    ) -> List[Features]:
        return [
            Features(
                scipy.sparse.coo_matrix(feature.features.sum(0)),
                FEATURE_TYPE_SENTENCE,
                feature.attribute,
                feature.origin,
            )
            for feature in sparse_sequence_features
        ]

    @staticmethod
    def _get_name_attribute(attributes: Set[Text]) -> Optional[Text]:
        # there is always either INTENT or ACTION_NAME
        return next(
            (
                attribute
                for attribute in attributes
                if attribute in {INTENT, ACTION_NAME}
            ),
            None,
        )

    def _extract_state_features(
        self,
        sub_state: SubState,
        precomputations: Optional[MessageContainerForCoreFeaturization],
        sparse: bool = False,
    ) -> Dict[Text, List[Features]]:
        # Remove entities from possible attributes
        attributes = set(
            attribute for attribute in sub_state.keys() if attribute != ENTITIES
        )

        if precomputations is not None:
            # Collect features for all those attributes
            attributes_to_features = precomputations.collect_features(
                sub_state, attributes=attributes
            )
            # if features for INTENT or ACTION_NAME exist,
            # they are always sparse sequence features;
            # transform them to sentence sparse features
            if attributes_to_features.get(INTENT):
                attributes_to_features[INTENT] = self._to_sparse_sentence_features(
                    attributes_to_features[INTENT]
                )
            if attributes_to_features.get(ACTION_NAME):
                attributes_to_features[ACTION_NAME] = self._to_sparse_sentence_features(
                    attributes_to_features[ACTION_NAME]
                )

            # Combine and sort the features:
            # Per attribute, combine features of same type and level into one Feature,
            # and (if there are any such features) store the results in a list where
            # - all the sparse features are listed first and a
            # - sequence feature is always listed before the sentence feature of the
            #   same type (sparse/not sparse).
            output = {
                attribute: Features.reduce(
                    features_list=features_list, expected_origins=None
                )
                for attribute, features_list in attributes_to_features.items()
                if len(features_list) > 0  # otherwise, following will fail
            }
        else:
            output = {}

        # Check that the name attribute has features
        name_attribute = self._get_name_attribute(attributes)
        if name_attribute and name_attribute not in output:
            # nlu pipeline didn't create features for user or action
            # this might happen, for example, when we have action_name in the state
            # but it did not get featurized because only character level
            # CountVectorsFeaturizer was included in the config.
            output[name_attribute] = self._create_features(
                sub_state, name_attribute, sparse
            )
        return output

    def encode_state(
        self,
        state: State,
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> Dict[Text, List[Features]]:
        """Encode the given state.

        Args:
            state: The state to encode
            precomputations: Contains precomputed features and attributes.

        Returns:
            A dictionary of state_type to list of features.
        """
        state_features = {}
        for state_type, sub_state in state.items():
            if state_type == PREVIOUS_ACTION:
                state_features.update(
                    self._extract_state_features(
                        sub_state, precomputations=precomputations, sparse=True
                    )
                )
            # featurize user only if it is "real" user input,
            # i.e. input from a turn after action_listen
            if state_type == USER and is_prev_action_listen_in_state(state):
                state_features.update(
                    self._extract_state_features(
                        sub_state, precomputations=precomputations, sparse=True
                    )
                )
                if sub_state.get(ENTITIES):
                    state_features[ENTITIES] = self._create_features(
                        sub_state, ENTITIES, sparse=True
                    )

            if state_type in {SLOTS, ACTIVE_LOOP}:
                state_features[state_type] = self._create_features(
                    sub_state, state_type, sparse=True
                )

        return state_features

    def encode_entities(
        self,
        entity_data: Dict[Text, Any],
        precomputations: Optional[MessageContainerForCoreFeaturization],
        bilou_tagging: bool = False,
    ) -> Dict[Text, List[Features]]:
        """Encode the given entity data.

        Produce numeric entity tags for tokens.

        Args:
            entity_data: The dict containing the text and entity labels and locations
            precomputations: Contains precomputed features and attributes.
            bilou_tagging: indicates whether BILOU tagging should be used or not

        Returns:
            A dictionary of entity type to list of features.
        """
        # TODO
        #  The entity states used to create the tag-idx-mapping contains the
        #  entities and the concatenated entity and roles/groups. We do not
        #  distinguish between entities and roles/groups right now.
        if (
            not entity_data
            or not self.entity_tag_specs
            or self.entity_tag_specs[0].num_tags < 2
        ):
            # we cannot build a classifier with fewer than 2 classes
            return {}
        if precomputations is None:
            message = None
        else:
            message = precomputations.lookup_message(user_text=entity_data[TEXT])
            message.data[ENTITIES] = entity_data[ENTITIES]

        if not message:
            return {}

        if bilou_tagging:
            bilou_utils.apply_bilou_schema_to_message(message)

        return {
            ENTITY_TAGS: [
                model_data_utils.get_tag_ids(
                    message, self.entity_tag_specs[0], bilou_tagging
                )
            ]
        }

    def _encode_action(
        self,
        action: Text,
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> Dict[Text, List[Features]]:
        if action in self.action_texts:
            action_as_sub_state = {ACTION_TEXT: action}
        else:
            action_as_sub_state = {ACTION_NAME: action}

        return self._extract_state_features(
            action_as_sub_state, precomputations=precomputations
        )

    def encode_all_labels(
        self,
        domain: Domain,
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> List[Dict[Text, List[Features]]]:
        """Encode all action from the domain.

        Args:
            domain: The domain that contains the actions.
            precomputations: Contains precomputed features and attributes.

        Returns:
            A list of encoded actions.
        """
        return [
            self._encode_action(action, precomputations)
            for action in domain.action_names_or_texts
        ]

    def to_dict(self) -> Dict[str, Any]:
        return {
            "action_texts": self.action_texts,
            "entity_tag_specs": self.entity_tag_specs,
            "feature_states": self._default_feature_states,
        }

    @classmethod
    def create_from_dict(
        cls, data: Dict[str, Any]
    ) -> Optional["SingleStateFeaturizer"]:
        if not data:
            return None

        featurizer = SingleStateFeaturizer()
        featurizer.action_texts = data["action_texts"]
        featurizer._default_feature_states = data["feature_states"]
        featurizer.entity_tag_specs = data["entity_tag_specs"]
        return featurizer


class IntentTokenizerSingleStateFeaturizer(SingleStateFeaturizer):
    """A SingleStateFeaturizer for use with policies that predict intent labels."""

    def _encode_intent(
        self,
        intent: Text,
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> Dict[Text, List[Features]]:
        """Extracts a numeric representation of an intent.

        Args:
            intent: Intent to be encoded.
            precomputations: Contains precomputed features and attributes.

        Returns:
            Encoded representation of intent.
        """
        intent_as_sub_state = {INTENT: intent}
        return self._extract_state_features(intent_as_sub_state, precomputations)

    def encode_all_labels(
        self,
        domain: Domain,
        precomputations: Optional[MessageContainerForCoreFeaturization],
    ) -> List[Dict[Text, List[Features]]]:
        """Encodes all relevant labels from the domain using the given precomputations.

        Args:
            domain: The domain that contains the labels.
            precomputations: Contains precomputed features and attributes.

        Returns:
            A list of encoded labels.
        """
        return [
            self._encode_intent(intent, precomputations) for intent in domain.intents
        ]



import asyncio
import functools
import importlib
import inspect
import logging
import pkgutil
import sys
from types import ModuleType
from typing import Text, Dict, Optional, Any, List, Callable, Collection, Type

import rasa.shared.utils.io
from rasa.shared.constants import DOCS_URL_MIGRATION_GUIDE
from rasa.shared.exceptions import RasaException

logger = logging.getLogger(__name__)


def class_from_module_path(
    module_path: Text, lookup_path: Optional[Text] = None
) -> Type:
    """Given the module name and path of a class, tries to retrieve the class.

    The loaded class can be used to instantiate new objects.

    Args:
        module_path: either an absolute path to a Python class,
                     or the name of the class in the local / global scope.
        lookup_path: a path where to load the class from, if it cannot
                     be found in the local / global scope.

    Returns:
        a Python class

    Raises:
        ImportError, in case the Python class cannot be found.
        RasaException, in case the imported result is something other than a class
    """
    warn_and_exit_if_module_path_contains_rasa_plus(module_path, lookup_path)

    klass = None
    if "." in module_path:
        module_name, _, class_name = module_path.rpartition(".")
        m = importlib.import_module(module_name)
        klass = getattr(m, class_name, None)
    elif lookup_path:
        # try to import the class from the lookup path
        m = importlib.import_module(lookup_path)
        klass = getattr(m, module_path, None)

    if klass is None:
        raise ImportError(f"Cannot retrieve class from path {module_path}.")

    if not inspect.isclass(klass):
        raise RasaException(
            f"`class_from_module_path()` is expected to return a class, "
            f"but for {module_path} we got a {type(klass)}."
        )
    return klass


def import_package_modules(package_name: str) -> List[ModuleType]:
    """Import all modules in a package."""
    package = importlib.import_module(package_name)
    return [
        importlib.import_module(f"{package_name}.{module_name}")
        for _, module_name, _ in pkgutil.iter_modules(package.__path__)
    ]


def all_subclasses(cls: Any) -> List[Any]:
    """Returns all known (imported) subclasses of a class."""
    classes = cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in all_subclasses(s)
    ]

    return [subclass for subclass in classes if not inspect.isabstract(subclass)]


def module_path_from_instance(inst: Any) -> Text:
    """Return the module path of an instance's class."""
    return inst.__module__ + "." + inst.__class__.__name__


def sort_list_of_dicts_by_first_key(dicts: List[Dict]) -> List[Dict]:
    """Sorts a list of dictionaries by their first key."""
    return sorted(dicts, key=lambda d: next(iter(d.keys())))


def lazy_property(function: Callable) -> Any:
    """Allows to avoid recomputing a property over and over.

    The result gets stored in a local var. Computation of the property
    will happen once, on the first call of the property. All
    succeeding calls will use the value stored in the private property.
    """
    attr_name = "_lazy_" + function.__name__

    def _lazyprop(self: Any) -> Any:
        if not hasattr(self, attr_name):
            setattr(self, attr_name, function(self))
        return getattr(self, attr_name)

    return property(_lazyprop)


def cached_method(f: Callable[..., Any]) -> Callable[..., Any]:
    """Caches method calls based on the call's `args` and `kwargs`.

    Works for `async` and `sync` methods. Don't apply this to functions.

    Args:
        f: The decorated method whose return value should be cached.

    Returns:
        The return value which the method gives for the first call with the given
        arguments.
    """
    assert "self" in arguments_of(f), "This decorator can only be used with methods."

    class Cache:
        """Helper class to abstract the caching details."""

        def __init__(self, caching_object: object, args: Any, kwargs: Any) -> None:
            self.caching_object = caching_object
            self.cache = getattr(caching_object, self._cache_name(), {})
            # noinspection PyUnresolvedReferences
            self.cache_key = functools._make_key(args, kwargs, typed=False)

        def _cache_name(self) -> Text:
            return f"_cached_{self.caching_object.__class__.__name__}_{f.__name__}"

        def is_cached(self) -> bool:
            return self.cache_key in self.cache

        def cache_result(self, result: Any) -> None:
            self.cache[self.cache_key] = result
            setattr(self.caching_object, self._cache_name(), self.cache)

        def cached_result(self) -> Any:
            return self.cache[self.cache_key]

    if asyncio.iscoroutinefunction(f):

        @functools.wraps(f)
        async def decorated(self: object, *args: Any, **kwargs: Any) -> Any:
            cache = Cache(self, args, kwargs)
            if not cache.is_cached():
                # Store the task immediately so that other concurrent calls of the
                # method can re-use the same task and don't schedule a second execution.
                to_cache = asyncio.ensure_future(f(self, *args, **kwargs))
                cache.cache_result(to_cache)
            return await cache.cached_result()

        return decorated
    else:

        @functools.wraps(f)
        def decorated(self: object, *args: Any, **kwargs: Any) -> Any:
            cache = Cache(self, args, kwargs)
            if not cache.is_cached():
                to_cache = f(self, *args, **kwargs)
                cache.cache_result(to_cache)
            return cache.cached_result()

        return decorated


def transform_collection_to_sentence(collection: Collection[Text]) -> Text:
    """Transforms e.g. a list like ['A', 'B', 'C'] into a sentence 'A, B and C'."""
    x = list(collection)
    if len(x) >= 2:
        return ", ".join(map(str, x[:-1])) + " and " + x[-1]
    return "".join(collection)


def minimal_kwargs(
    kwargs: Dict[Text, Any], func: Callable, excluded_keys: Optional[List] = None
) -> Dict[Text, Any]:
    """Returns only the kwargs which are required by a function. Keys, contained in
    the exception list, are not included.

    Args:
        kwargs: All available kwargs.
        func: The function which should be called.
        excluded_keys: Keys to exclude from the result.

    Returns:
        Subset of kwargs which are accepted by `func`.

    """
    excluded_keys = excluded_keys or []

    possible_arguments = arguments_of(func)

    return {
        k: v
        for k, v in kwargs.items()
        if k in possible_arguments and k not in excluded_keys
    }


def mark_as_experimental_feature(feature_name: Text) -> None:
    """Warns users that they are using an experimental feature."""
    logger.warning(
        f"The {feature_name} is currently experimental and might change or be "
        "removed in the future 🔬 Please share your feedback on it in the "
        "forum (https://forum.rasa.com) to help us make this feature "
        "ready for production."
    )


def arguments_of(func: Callable) -> List[Text]:
    """Return the parameters of the function `func` as a list of names."""
    import inspect

    return list(inspect.signature(func).parameters.keys())


def extract_duplicates(list1: List[Any], list2: List[Any]) -> List[Any]:
    """Extracts duplicates from two lists."""
    if list1:
        dict1 = {
            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in list1
        }
    else:
        dict1 = {}

    if list2:
        dict2 = {
            (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in list2
        }
    else:
        dict2 = {}

    set1 = set(dict1.keys())
    set2 = set(dict2.keys())
    dupes = set1.intersection(set2)
    return sorted(list(dupes))


def clean_duplicates(dupes: Dict[Text, Any]) -> Dict[Text, Any]:
    """Removes keys for empty values."""
    duplicates = dupes.copy()
    for k in dupes:
        if not dupes[k]:
            duplicates.pop(k)

    return duplicates


def merge_dicts(
    tempDict1: Dict[Text, Any],
    tempDict2: Dict[Text, Any],
    override_existing_values: bool = False,
) -> Dict[Text, Any]:
    """Merges two dicts."""
    if override_existing_values:
        merged_dicts, b = tempDict1.copy(), tempDict2.copy()

    else:
        merged_dicts, b = tempDict2.copy(), tempDict1.copy()
    merged_dicts.update(b)
    return merged_dicts


def merge_lists(
    list1: List[Any], list2: List[Any], override: bool = False
) -> List[Any]:
    """Merges two lists."""
    return sorted(list(set(list1 + list2)))


def merge_lists_of_dicts(
    dict_list1: List[Dict],
    dict_list2: List[Dict],
    override_existing_values: bool = False,
) -> List[Dict]:
    """Merges two dict lists."""
    dict1 = {
        (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in dict_list1
    }
    dict2 = {
        (sorted(list(i.keys()))[0] if isinstance(i, dict) else i): i for i in dict_list2
    }
    merged_dicts = merge_dicts(dict1, dict2, override_existing_values)
    return list(merged_dicts.values())


def warn_and_exit_if_module_path_contains_rasa_plus(
    module_path: Text, lookup_path: Optional[str] = None
) -> None:
    """Warns and exits if the module path contains `rasa_plus`.

    Args:
        module_path: The module path to check.
        lookup_path: The lookup path to check.
    """
    if "rasa_plus" in module_path.lower() or (
        lookup_path and "rasa_plus" in lookup_path.lower()
    ):
        rasa.shared.utils.io.raise_warning(
            f"Your module path '{module_path}' contains 'rasa_plus'. "
            f"The path to this Rasa Pro component has changed, please "
            f"follow the migration guide in the official documentation "
            f"to update your code: ",
            docs=DOCS_URL_MIGRATION_GUIDE,
        )
        sys.exit(1)



import logging
from pathlib import Path
from typing import Optional, Dict, Text, List, Any, Union

from rasa.shared.core.domain import Domain
from rasa.shared.core.events import SlotSet, ActionExecuted, Event
from rasa.shared.exceptions import RasaCoreException
from rasa.shared.core.training_data.story_reader.story_step_builder import (
    StoryStepBuilder,
)
from rasa.shared.core.training_data.structures import StoryStep

logger = logging.getLogger(__name__)


class StoryReader:
    """Helper class to read a story file."""

    def __init__(
        self, domain: Optional[Domain] = None, source_name: Optional[Text] = None
    ) -> None:
        """Constructor for the StoryReader.

        Args:
            domain: Domain object.
            source_name: Name of the training data source.
        """
        self.story_steps: List[StoryStep] = []
        self.current_step_builder: Optional[StoryStepBuilder] = None
        self.domain = domain
        self.source_name = source_name
        self._is_parsing_conditions = False

    def read_from_file(
        self, filename: Text, skip_validation: bool = False
    ) -> List[StoryStep]:
        """Reads stories or rules from file.

        Args:
            filename: Path to the story/rule file.
            skip_validation: `True` if file validation should be skipped.

        Returns:
            `StoryStep`s read from `filename`.
        """
        raise NotImplementedError

    @staticmethod
    def is_stories_file(filename: Union[Text, Path]) -> bool:
        """Checks if the specified file is a story file.

        Args:
            filename: File to check.

        Returns:
            `True` if specified file is a story file, `False` otherwise.
        """
        raise NotImplementedError

    def _add_current_stories_to_result(self) -> None:
        if self.current_step_builder:
            self.current_step_builder.flush()
            self.story_steps.extend(self.current_step_builder.story_steps)

    def _new_story_part(self, name: Text, source_name: Optional[Text]) -> None:
        self._add_current_stories_to_result()
        self.current_step_builder = StoryStepBuilder(name, source_name)

    def _new_rule_part(self, name: Text, source_name: Optional[Text]) -> None:
        self._add_current_stories_to_result()
        self.current_step_builder = StoryStepBuilder(name, source_name, is_rule=True)

    def _parse_events(
        self, event_name: Text, parameters: Dict[Text, Any]
    ) -> Optional[List["Event"]]:
        # add 'name' only if event is not a SlotSet,
        # because there might be a slot with slot_key='name'
        if "name" not in parameters and event_name != SlotSet.type_name:
            parameters["name"] = event_name

        parsed_events = Event.from_story_string(
            event_name, parameters, default=ActionExecuted
        )
        if parsed_events is None:
            raise StoryParseError(
                "Unknown event '{}'. It is Neither an event nor an action).".format(
                    event_name
                )
            )

        return parsed_events

    def _add_event(self, event_name: Text, parameters: Dict[Text, Any]) -> None:
        parsed_events = self._parse_events(event_name, parameters)
        if parsed_events is None:
            parsed_events = []

        if self.current_step_builder is None:
            raise StoryParseError(
                "Failed to handle event '{}'. There is no "
                "started story block available. "
                "".format(event_name)
            )

        for p in parsed_events:
            if self._is_parsing_conditions:
                self.current_step_builder.add_event_as_condition(p)
            else:
                self.current_step_builder.add_event(p)

    def _add_checkpoint(
        self, name: Text, conditions: Optional[Dict[Text, Any]]
    ) -> None:
        # Ensure story part already has a name
        if not self.current_step_builder:
            raise StoryParseError(
                "Checkpoint '{}' is at an invalid location. "
                "Expected a story start.".format(name)
            )

        self.current_step_builder.add_checkpoint(name, conditions)


class StoryParseError(RasaCoreException, ValueError):
    """Raised if there is an error while parsing a story file."""

    def __init__(self, message: Text) -> None:
        self.message = message
        super(StoryParseError, self).__init__()



import logging
from pathlib import Path
from typing import Any, Dict, List, Text

import requests
from rasa.shared.core.domain import Domain
from rasa.shared.utils.cli import print_error_and_exit

from rasa.markers.validate import validate_marker_file
from rasa.utils.yaml import (
    YAML_CONFIGS,
    collect_configs_from_yaml_files,
    collect_yaml_files_from_path,
)

PATTERNS_PATH = "/api/v1/patterns"

log = logging.getLogger(__name__)


def upload(url: Text, domain: Domain, markers_path: Path) -> None:
    validate_marker_file(domain, markers_path)
    yaml_files = collect_yaml_files_from_path(markers_path)
    yaml_configs = collect_configs_from_yaml_files(yaml_files)

    marker_json_list = _convert_yaml_to_json(yaml_configs)

    patterns_url = f"{url}{PATTERNS_PATH}"
    try:
        response = requests.post(f"{patterns_url}", json={"patterns": marker_json_list})
    except requests.exceptions.ConnectionError:
        print_error_and_exit(
            f"Failed to connect to Rasa Pro Services at {patterns_url}. "
            f"Make sure the server is running and the correct URL is configured."
        )
        # this shouldn't be required but I need to add this for the test
        # test_upload::test_upload_with_connection_error to pass
        return

    if response.status_code == 200:
        patterns = response.json()
        log.info(
            f"Successfully uploaded markers to {url}"
            f"Total count: {patterns.get('count')}\n"
            f"Updated: {patterns.get('updated')}\n"
            f"Inserted: {patterns.get('inserted')}\n"
            f"Deleted: {patterns.get('deleted')}\n"
        )
    else:
        print_error_and_exit(
            f"Failed to upload markers to {patterns_url}. "
            f"Status Code: {response.status_code} "
            f"Response: {response.text}"
        )


def _convert_marker_config_to_json(
    marker_pattern_name: Text, config: Dict[Text, Any]
) -> Dict[Text, Any]:
    description = config.pop("description", None)
    return {
        "name": marker_pattern_name,
        "description": description,
        "config": config,
    }


def _convert_yaml_to_json(yaml_configs: YAML_CONFIGS) -> List[Dict]:
    marker_json_list: List[Dict] = []
    for _, yaml_config in yaml_configs.items():
        for marker_pattern_name, config in yaml_config.items():
            marker_json = _convert_marker_config_to_json(marker_pattern_name, config)
            marker_json_list.append(marker_json)
    return marker_json_list



import argparse
from rasa.cli.arguments import default_arguments
from rasa.cli.arguments.run import add_server_arguments


def set_x_arguments(parser: argparse.ArgumentParser) -> None:
    """Add CLI arguments for running rasa x --production."""
    parser.add_argument(
        "--production",
        action="store_true",
        help="Run a Rasa server in a mode that allows connecting "
        "to Rasa Enterprise as the config endpoint",
    )

    parser.add_argument(
        "--config-endpoint",
        type=str,
        help="Rasa X endpoint URL from which to pull the runtime config. This URL "
        "typically contains the Rasa X token for authentication. Example: "
        "https://example.com/api/config?token=my_rasa_x_token",
    )

    parser.add_argument(
        "--no-prompt",
        action="store_true",
        help="Automatic yes or default options to prompts and oppressed warnings.",
    )

    default_arguments.add_model_param(parser, add_positional_arg=False)
    add_server_arguments(parser)



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any, Dict, Optional, Text

from rasa.utils.common import read_global_config_value, write_global_config_value

from rasa.studio.constants import (
    RASA_STUDIO_AUTH_SERVER_URL_ENV,
    RASA_STUDIO_CLI_CLIENT_ID_KEY_ENV,
    RASA_STUDIO_CLI_DISABLE_VERIFY_KEY_ENV,
    RASA_STUDIO_CLI_REALM_NAME_KEY_ENV,
    RASA_STUDIO_CLI_STUDIO_URL_ENV,
    STUDIO_CONFIG_KEY,
)

AUTH_SERVER_URL_KEY = "authentication_server_url"
STUDIO_URL_KEY = "studio_url"
CLIENT_ID_KEY = "client_id"
REALM_NAME_KEY = "realm_name"
CLIENT_SECRET_KEY = "client_secret"
DISABLE_VERIFY = "disable_verify"


@dataclass
class StudioConfig:
    authentication_server_url: Optional[Text]
    studio_url: Optional[Text]
    client_id: Optional[Text]
    realm_name: Optional[Text]
    disable_verify: bool = False

    def to_dict(self) -> Dict[Text, Optional[Any]]:
        return {
            AUTH_SERVER_URL_KEY: self.authentication_server_url,
            STUDIO_URL_KEY: self.studio_url,
            CLIENT_ID_KEY: self.client_id,
            REALM_NAME_KEY: self.realm_name,
            DISABLE_VERIFY: self.disable_verify,
        }

    @classmethod
    def from_dict(cls, data: Dict[Text, Text]) -> StudioConfig:
        return cls(
            authentication_server_url=data[AUTH_SERVER_URL_KEY],
            studio_url=data[STUDIO_URL_KEY],
            client_id=data[CLIENT_ID_KEY],
            realm_name=data[REALM_NAME_KEY],
            disable_verify=data.get(DISABLE_VERIFY, False),
        )

    def write_config(self) -> None:
        write_global_config_value(
            STUDIO_CONFIG_KEY,
            self.to_dict(),
        )

    @staticmethod
    def read_config() -> StudioConfig:
        env_config = StudioConfig._read_env_config()
        file_config = StudioConfig._read_config_from_file()

        return env_config._merge(file_config)

    def is_valid(self) -> bool:
        return all(
            [
                self.authentication_server_url,
                self.studio_url,
                self.client_id,
                self.realm_name,
            ]
        )

    @staticmethod
    def _read_config_from_file() -> StudioConfig:
        config = read_global_config_value(STUDIO_CONFIG_KEY, unavailable_ok=True)

        if config is None:
            return StudioConfig(None, None, None, None, False)

        if not isinstance(config, dict):
            raise ValueError(
                "Invalid config file format. "
                "Expected a dictionary, but found a {}."
                "".format(type(config).__name__)
            )

        for key in config:
            if not isinstance(config[key], str) and key != DISABLE_VERIFY:
                raise ValueError(
                    "Invalid config file format. "
                    f"Key '{key}' is not a text value."
                    "Please provide a valid value for the key."
                    ""
                )

        return StudioConfig.from_dict(config)

    @staticmethod
    def _read_env_config() -> StudioConfig:
        return StudioConfig(
            authentication_server_url=StudioConfig._read_env_value(
                RASA_STUDIO_AUTH_SERVER_URL_ENV
            ),
            studio_url=StudioConfig._read_env_value(RASA_STUDIO_CLI_STUDIO_URL_ENV),
            client_id=StudioConfig._read_env_value(RASA_STUDIO_CLI_CLIENT_ID_KEY_ENV),
            realm_name=StudioConfig._read_env_value(RASA_STUDIO_CLI_REALM_NAME_KEY_ENV),
            disable_verify=bool(
                os.getenv(RASA_STUDIO_CLI_DISABLE_VERIFY_KEY_ENV, False)
            ),
        )

    @staticmethod
    def _read_env_value(env_name: Text) -> Optional[Text]:
        value = os.getenv(env_name, None)
        if value == "":
            raise ValueError(
                "Invalid config file format. "
                f"Key '{env_name}' is not a text value."
                "Please provide a valid value for the key."
                ""
            )
        return value

    def _merge(self, other: StudioConfig) -> StudioConfig:
        return StudioConfig(
            authentication_server_url=(
                self.authentication_server_url or other.authentication_server_url
            ),
            studio_url=self.studio_url or other.studio_url,
            client_id=self.client_id or other.client_id,
            realm_name=self.realm_name or other.realm_name,
            disable_verify=self.disable_verify or other.disable_verify,
        )



import json
import logging
from typing import Any, Dict, Optional, TYPE_CHECKING

import aiohttp

from rasa.core.actions.action_exceptions import ActionExecutionRejection, DomainNotFound
from rasa.core.actions.custom_action_executor import (
    CustomActionExecutor,
    CustomActionRequestWriter,
)
from rasa.core.constants import (
    COMPRESS_ACTION_SERVER_REQUEST_ENV_NAME,
    DEFAULT_COMPRESS_ACTION_SERVER_REQUEST,
    DEFAULT_REQUEST_TIMEOUT,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.utils.common import get_bool_env_variable

if TYPE_CHECKING:
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.trackers import DialogueStateTracker

from rasa.utils.endpoints import ClientResponseError, EndpointConfig

logger = logging.getLogger(__name__)


class HTTPCustomActionExecutor(CustomActionExecutor):
    """HTTP-based implementation of the CustomActionExecutor.

    Executes custom actions by making HTTP POST requests to the action endpoint.
    """

    def __init__(
        self,
        action_name: str,
        action_endpoint: EndpointConfig,
    ) -> None:
        self.action_name = action_name
        self.action_endpoint = action_endpoint
        self.request_writer = CustomActionRequestWriter(action_name, action_endpoint)
        self.should_compress = get_bool_env_variable(
            COMPRESS_ACTION_SERVER_REQUEST_ENV_NAME,
            DEFAULT_COMPRESS_ACTION_SERVER_REQUEST,
        )

    async def run(
        self,
        tracker: "DialogueStateTracker",
        domain: Optional["Domain"] = None,
        include_domain: bool = False,
    ) -> Dict[str, Any]:
        """Execute the custom action using an HTTP POST request.

        Args:
            tracker: The current state of the dialogue.
            domain: The domain object containing domain-specific information.
            include_domain: If True, the domain is included in the request.

        Returns:
            A dictionary containing the response from the custom action endpoint.

        Raises:
            RasaException: If an error occurs while making the HTTP request.
        """
        try:
            logger.debug(
                "Calling action endpoint to run action '{}'.".format(self.action_name)
            )

            json_body = self.request_writer.create(
                tracker=tracker, domain=domain, include_domain=include_domain
            )

            response = await self._perform_request_with_retries(json_body)

            if response is None:
                response = {}

            return response

        except ClientResponseError as e:
            if e.status == 400:
                response_data = json.loads(e.text)
                exception = ActionExecutionRejection(
                    response_data["action_name"], response_data.get("error")
                )
                logger.error(exception.message)
                raise exception
            else:
                raise RasaException(
                    f"Failed to execute custom action '{self.action_name}'"
                ) from e

        except aiohttp.ClientConnectionError as e:
            logger.error(
                f"Failed to run custom action '{self.action_name}'. Couldn't connect "
                f"to the server at '{self.action_endpoint.url}'. "
                f"Is the server running? "
                f"Error: {e}"
            )
            raise RasaException(
                f"Failed to execute custom action '{self.action_name}'. "
                f"Couldn't connect to the server at '{self.action_endpoint.url}."
            )

        except aiohttp.ClientError as e:
            # not all errors have a status attribute, but
            # helpful to log if they got it

            # noinspection PyUnresolvedReferences
            status = getattr(e, "status", None)
            raise RasaException(
                "Failed to run custom action '{}'. Action server "
                "responded with a non 200 status code of {}. "
                "Make sure your action server properly runs actions "
                "and returns a 200 once the action is executed. "
                "Error: {}".format(self.action_name, status, e)
            )

    async def _perform_request_with_retries(
        self,
        json_body: Dict[str, Any],
    ) -> Any:
        """Attempts to perform the request with retries if necessary."""
        assert self.action_endpoint is not None
        try:
            return await self.action_endpoint.request(
                json=json_body,
                method="post",
                timeout=DEFAULT_REQUEST_TIMEOUT,
                compress=self.should_compress,
            )
        except ClientResponseError as e:
            # Repeat the request because Domain was not in the payload
            if e.status == 449:
                raise DomainNotFound()
            raise e



from typing import Any, Dict, List, Optional
import os
import structlog

from rasa.shared.constants import (
    AZURE_API_BASE_ENV_VAR,
    AZURE_API_KEY_ENV_VAR,
    AZURE_API_TYPE_ENV_VAR,
    AZURE_API_VERSION_ENV_VAR,
    OPENAI_API_BASE_ENV_VAR,
    OPENAI_API_KEY_ENV_VAR,
    OPENAI_API_TYPE_ENV_VAR,
    OPENAI_API_VERSION_ENV_VAR,
    API_BASE_CONFIG_KEY,
    API_KEY,
    API_VERSION_CONFIG_KEY,
    AZURE_OPENAI_PROVIDER,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers._configs.azure_openai_client_config import (
    AzureOpenAIClientConfig,
)
from rasa.shared.providers.embedding._base_litellm_embedding_client import (
    _BaseLiteLLMEmbeddingClient,
)
from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()


class AzureOpenAIEmbeddingClient(_BaseLiteLLMEmbeddingClient):
    """A client for interfacing with Azure's OpenAI embedding deployments.

    Parameters:
        deployment (str): The deployment name.
        model (Optional[str]): The embedding model name.
        api_base (Optional[str]): The base URL for the API endpoints.
            If not provided, it will be set via environment variable.
        api_type (Optional[str]): The type of the API to use.
            If not provided, it will be set via environment variable.
        api_version (Optional[str]): The version of the API to use.
            If not provided, it will be set via environment variable.
        kwargs (Optional[Dict[str, Any]]): Optional configuration parameters specific
            to the embedding model deployment.
    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
        DeprecationWarning: If deprecated environment variables are used for
            configuration.
    """

    def __init__(
        self,
        deployment: str,
        model: Optional[str] = None,
        api_base: Optional[str] = None,
        api_type: Optional[str] = None,
        api_version: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._deployment = deployment
        self._model = model

        # Set api_base with the following priority:
        # parameter -> Azure Env Var -> (deprecated) OpenAI Env Var
        self._api_base = (
            api_base
            or os.environ.get(AZURE_API_BASE_ENV_VAR)
            or os.environ.get(OPENAI_API_BASE_ENV_VAR)
        )
        self._api_type = (
            api_type
            or os.environ.get(AZURE_API_TYPE_ENV_VAR)
            or os.environ.get(OPENAI_API_TYPE_ENV_VAR)
        )
        self._api_version = (
            api_version
            or os.environ.get(AZURE_API_VERSION_ENV_VAR)
            or os.environ.get(OPENAI_API_VERSION_ENV_VAR)
        )
        # Litellm does not support use of OPENAI_API_KEY, so we need to map it
        # because of backward compatibility. However, we're first looking at
        # AZURE_API_KEY.
        self._api_key = os.environ.get(AZURE_API_KEY_ENV_VAR) or os.environ.get(
            OPENAI_API_KEY_ENV_VAR
        )

        self._extra_parameters = kwargs or {}
        self.validate_client_setup()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "AzureOpenAIEmbeddingClient":
        """
        Initializes the client from given configuration.

        Args:
            config (Dict[str, Any]): Configuration.

        Raises:
            ValueError: Raised in cases of invalid configuration:
                - If any of the required configuration keys are missing.
                - If `api_type` has a value different from `azure`.

        Returns:
            AzureOpenAIEmbeddingClient: Initialized client.
        """
        try:
            azure_openai_config = AzureOpenAIClientConfig.from_dict(config)
        except ValueError as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "azure_openai_embedding_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            deployment=azure_openai_config.deployment,
            model=azure_openai_config.model,
            api_base=azure_openai_config.api_base,
            api_type=azure_openai_config.api_type,
            api_version=azure_openai_config.api_version,
            **azure_openai_config.extra_parameters,
        )

    @property
    def config(self) -> dict:
        """Returns the configuration for that the llm client
        in dictionary form.
        """
        config = AzureOpenAIClientConfig(
            deployment=self.deployment,
            model=self.model,
            api_base=self.api_base,
            api_type=self.api_type,
            api_version=self.api_version,
            extra_parameters=self._extra_parameters,
        )
        return config.to_dict()

    @property
    def model(self) -> Optional[str]:
        """
        Returns the name of the model deployed on Azure. If model name is not
        provided, returns "N/A".
        """
        return self._model

    @property
    def deployment(self) -> str:
        """Return the deployment name for the azure openai embedding client."""
        return self._deployment

    @property
    def api_base(self) -> Optional[str]:
        """Returns the base API URL for the oazure penai embedding client."""
        return self._api_base

    @property
    def api_type(self) -> Optional[str]:
        """Returns the API type for the azure openai embedding client."""
        return self._api_type

    @property
    def api_version(self) -> Optional[str]:
        """Returns the API version for the azure openai embedding client."""
        return self._api_version

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        """
        Returns the model parameters for the azure openai embedding client.

        Returns:
            Dictionary containing the model parameters.
        """
        return self._extra_parameters

    @property
    def _embedding_fn_args(self) -> dict:
        return {
            **self._litellm_extra_parameters,
            "model": self._litellm_model_name,
            "api_base": self.api_base,
            "api_type": self.api_type,
            "api_version": self.api_version,
            "api_key": self._api_key,
        }

    @property
    def _litellm_model_name(self) -> str:
        """Get the model name formatted for azure openai embedding client."""
        if self.deployment and f"{AZURE_OPENAI_PROVIDER}/" not in self.deployment:
            return f"{AZURE_OPENAI_PROVIDER}/{self.deployment}"
        return self.deployment

    def validate_client_setup(self) -> None:
        """Perform client validation. By default only environment variables
        are validated.

        Raises:
            ProviderClientValidationError if validation fails.

        Throws:
            DeprecationWarning: If deprecated environment variables are used.
        """
        _substitute_env_key_mapper = [
            {
                "param_name": "API base",
                "config_key": API_BASE_CONFIG_KEY,
                "deprecated_env_key": OPENAI_API_BASE_ENV_VAR,
                "current_value": self.api_base,
                "new_env_key": AZURE_API_BASE_ENV_VAR,
            },
            {
                "param_name": "API key",
                "config_key": API_KEY,
                "deprecated_env_key": OPENAI_API_KEY_ENV_VAR,
                "current_value": self._api_key,
                "new_env_key": AZURE_API_KEY_ENV_VAR,
            },
            {
                "param_name": "API version",
                "config_key": API_VERSION_CONFIG_KEY,
                "deprecated_env_key": OPENAI_API_VERSION_ENV_VAR,
                "current_value": self.api_version,
                "new_env_key": AZURE_API_VERSION_ENV_VAR,
            },
        ]

        self._throw_deprecation_warnings(_substitute_env_key_mapper)
        self._validate_client_setup(_substitute_env_key_mapper)

    def _throw_deprecation_warnings(self, mappings: List[Dict[str, Any]]) -> None:
        """Throw deprecation warnings for deprecated environment variables."""
        for mapping in mappings:
            # Value was set through the new environment variable
            if os.environ.get(mapping["new_env_key"]) is not None and mapping[
                "current_value"
            ] == os.environ.get(mapping["new_env_key"]):
                continue

            # Value was set through the deprecated environment variable
            if mapping["current_value"] == os.environ.get(
                mapping["deprecated_env_key"]
            ):
                raise_deprecation_warning(
                    message=(
                        f"Usage of {mapping['deprecated_env_key']} environment "
                        f"variable for setting the {mapping['param_name']} of Azure "
                        f"OpenAI client is deprecated and will be removed in 4.0.0. "
                        f"Please use {mapping['new_env_key']} instead."
                    )
                )

    def _validate_client_setup(self, mappings: List[Dict[str, Any]]) -> None:
        """Validate environment variables for the client."""
        missing_environment_variable = False
        for mapping in mappings:
            if not mapping["current_value"]:
                event_info = (
                    f"Environment variable: {mapping['new_env_key']} or config key: "
                    f"{mapping['config_key']} is not set. Required for API calls. "
                )
                structlogger.error(
                    "azure_openai_embedding_client.validate_environment_variables",
                    event_info=event_info,
                    missing_environment_variable=mapping["new_env_key"],
                )
                missing_environment_variable = True

        if missing_environment_variable:
            raise ProviderClientValidationError(
                "Missing required environment variables/config keys for API calls."
            )



from __future__ import annotations

import abc
import logging
import os
from typing import Any, Dict, Optional, Text

import grpc
from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.metrics import set_meter_provider
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from rasa.core.actions.custom_action_executor import (
    CustomActionExecutor,
    RetryCustomActionExecutor,
)
from rasa.core.actions.grpc_custom_action_executor import GRPCCustomActionExecutor
from rasa.core.agent import Agent
from rasa.core.processor import MessageProcessor
from rasa.core.tracker_store import TrackerStore
from rasa.dialogue_understanding.commands import (
    Command,
    FreeFormAnswerCommand,
)
from rasa.dialogue_understanding.generator import (
    LLMCommandGenerator,
    SingleStepLLMCommandGenerator,
    MultiStepLLMCommandGenerator,
)
from rasa.dialogue_understanding.generator.nlu_command_adapter import NLUCommandAdapter
from rasa.engine.graph import GraphNode
from rasa.engine.training.graph_trainer import GraphTrainer
from rasa.tracing.constants import (
    ENDPOINTS_ENDPOINT_KEY,
    ENDPOINTS_INSECURE_KEY,
    ENDPOINTS_OTLP_BACKEND_TYPE,
    ENDPOINTS_TRACING_KEY,
    ENDPOINTS_TRACING_SERVICE_NAME_KEY,
    ENDPOINTS_ROOT_CERTIFICATES_KEY,
    ENDPOINTS_METRICS_KEY,
)
from rasa.tracing.instrumentation import instrumentation
from rasa.tracing.metric_instrument_provider import MetricInstrumentProvider
from rasa.utils.endpoints import EndpointConfig, read_endpoint_config

TRACING_SERVICE_NAME = os.environ.get("TRACING_SERVICE_NAME", "rasa")

logger = logging.getLogger(__name__)


def configure_tracing(tracer_provider: Optional[TracerProvider]) -> None:
    """Configure tracing functionality.

    When a tracing backend is defined, this function will
    instrument all methods that shall be traced.
    If no tracing backend is defined, no tracing is configured.

    :param tracer_provider: The `TracingProvider` to be used for tracing
    """
    if tracer_provider is None:
        return None

    from rasa.core.information_retrieval.information_retrieval import (
        InformationRetrieval,
    )
    from rasa.core.nlg.contextual_response_rephraser import ContextualResponseRephraser
    from rasa.core.policies.policy import Policy
    from rasa.engine.recipes.default_components import DEFAULT_COMPONENTS

    command_subclasses = [subclass for subclass in Command.__subclasses__()] + [
        subclass for subclass in FreeFormAnswerCommand.__subclasses__()
    ]

    policy_subclasses = [
        policy_class
        for policy_class in DEFAULT_COMPONENTS
        if issubclass(policy_class, Policy)
    ]

    vector_store_subclasses = [
        vector_store_class
        for vector_store_class in InformationRetrieval.__subclasses__()
    ]

    custom_action_executor_subclasses = []
    for custom_action_executor_class in CustomActionExecutor.__subclasses__():
        if custom_action_executor_class != RetryCustomActionExecutor:
            custom_action_executor_subclasses.append(custom_action_executor_class)

    instrumentation.instrument(
        tracer_provider=tracer_provider,
        agent_class=Agent,
        processor_class=MessageProcessor,
        tracker_store_class=TrackerStore,
        graph_node_class=GraphNode,
        graph_trainer_class=GraphTrainer,
        llm_command_generator_class=LLMCommandGenerator,
        command_subclasses=command_subclasses,
        contextual_response_rephraser_class=ContextualResponseRephraser,
        policy_subclasses=policy_subclasses,
        vector_store_subclasses=vector_store_subclasses,
        nlu_command_adapter_class=NLUCommandAdapter,
        endpoint_config_class=EndpointConfig,
        grpc_custom_action_executor_class=GRPCCustomActionExecutor,
        single_step_llm_command_generator_class=SingleStepLLMCommandGenerator,
        multi_step_llm_command_generator_class=MultiStepLLMCommandGenerator,
        custom_action_executor_subclasses=custom_action_executor_subclasses,
    )


def get_tracer_provider(endpoints_file: Text) -> Optional[TracerProvider]:
    """Configure tracing backend.

    When a known tracing backend is defined in the endpoints file, this
    function will configure the tracing infrastructure. When no or an unknown
    tracing backend is defined, this function does nothing.

    :param endpoints_file: The configuration file containing information about the
        tracing backend.
    :return: The `TracingProvider` to be used for all subsequent tracing.
    """
    cfg = read_endpoint_config(endpoints_file, ENDPOINTS_TRACING_KEY)

    if not cfg:
        logger.info(
            f"No endpoint for tracing type available in {endpoints_file},"
            f"tracing will not be configured."
        )
        return None
    if cfg.type == "jaeger":
        tracer_provider = JaegerTracerConfigurer.configure_from_endpoint_config(cfg)
    elif cfg.type == ENDPOINTS_OTLP_BACKEND_TYPE:
        tracer_provider = OTLPCollectorConfigurer.configure_from_endpoint_config(cfg)
    else:
        logger.warning(
            f"Unknown tracing type {cfg.type} read from {endpoints_file}, ignoring."
        )
        return None

    return tracer_provider


class TracerConfigurer(abc.ABC):
    """Abstract superclass for tracing configuration.

    `TracerConfigurer` is the abstract superclass from which all configurers
    for different supported backends should inherit.
    """

    @classmethod
    @abc.abstractmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing.

        This abstract method should be implemented by all concrete `TracerConfigurer`s.
        It shall read the configuration from the supplied argument, configure all
        necessary infrastructure for tracing, and return the `TracerProvider` to be
        used for tracing purposes.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """


class JaegerTracerConfigurer(TracerConfigurer):
    """The `TracerConfigurer` for a Jaeger backend."""

    @classmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing for Jaeger.

        This will read the Jaeger-specific configuration from the `EndpointConfig` and
        create a corresponding `TracerProvider` that exports to the given Jaeger
        backend.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """
        provider = TracerProvider(
            resource=Resource.create(
                {
                    SERVICE_NAME: cfg.kwargs.get(
                        ENDPOINTS_TRACING_SERVICE_NAME_KEY, TRACING_SERVICE_NAME
                    )
                }
            )
        )

        jaeger_exporter = JaegerExporter(
            **cls._extract_config(cfg), udp_split_oversized_batches=True
        )
        logger.info(
            f"Registered {cfg.type} endpoint for tracing. Traces will be exported to"
            f" {jaeger_exporter.agent_host_name}:{jaeger_exporter.agent_port}"
        )
        provider.add_span_processor(BatchSpanProcessor(jaeger_exporter))

        return provider

    @classmethod
    def _extract_config(cls, cfg: EndpointConfig) -> Dict[str, Any]:
        return {
            "agent_host_name": (cfg.kwargs.get("host", "localhost")),
            "agent_port": (cfg.kwargs.get("port", 6831)),
            "username": cfg.kwargs.get("username"),
            "password": cfg.kwargs.get("password"),
        }


class OTLPCollectorConfigurer(TracerConfigurer):
    """The `TracerConfigurer` for an OTLP collector backend."""

    @classmethod
    def configure_from_endpoint_config(cls, cfg: EndpointConfig) -> TracerProvider:
        """Configure tracing for OTLP Collector.

        This will read the OTLP collector-specific configuration from the
        `EndpointConfig` and create a corresponding `TracerProvider` that exports to
        the given OTLP collector.
        Currently, this only supports insecure connections via gRPC.

        :param cfg: The configuration to be read for configuring tracing.
        :return: The configured `TracerProvider`.
        """
        provider = TracerProvider(
            resource=Resource.create(
                {
                    SERVICE_NAME: cfg.kwargs.get(
                        ENDPOINTS_TRACING_SERVICE_NAME_KEY, TRACING_SERVICE_NAME
                    )
                }
            )
        )

        insecure = cfg.kwargs.get(ENDPOINTS_INSECURE_KEY)

        credentials = cls._get_credentials(cfg, insecure)

        otlp_exporter = OTLPSpanExporter(
            endpoint=cfg.kwargs[ENDPOINTS_ENDPOINT_KEY],
            insecure=insecure,
            credentials=credentials,
        )
        logger.info(
            f"Registered {cfg.type} endpoint for tracing."
            f"Traces will be exported to {cfg.kwargs['endpoint']}"
        )
        provider.add_span_processor(BatchSpanProcessor(otlp_exporter))

        return provider

    @classmethod
    def _get_credentials(
        cls, cfg: EndpointConfig, insecure: bool
    ) -> Optional[grpc.ChannelCredentials]:
        return _get_credentials(cfg, insecure)


class OTLPMetricConfigurer:
    """The metric configurer for the OTLP Collector backend."""

    @classmethod
    def configure_from_endpoint_config(
        cls, cfg: EndpointConfig
    ) -> Optional[OTLPMetricExporter]:
        """Configure metrics for OTLP Collector.

        This will read the OTLP collector-specific configuration from the
        `EndpointConfig` and create a corresponding `OTLPMetricExporter` that exports to
        the given OTLP collector.

        :param cfg: The configuration to be read for configuring metrics.
        :return: The configured `OTLPMetricExporter`.
        """
        insecure = cfg.kwargs.get(ENDPOINTS_INSECURE_KEY)

        credentials = cls._get_credentials(cfg, insecure)

        otlp_metric_exporter = OTLPMetricExporter(
            endpoint=cfg.kwargs[ENDPOINTS_ENDPOINT_KEY],
            insecure=insecure,
            credentials=credentials,
        )
        logger.info(
            f"Registered '{cfg.type}' endpoint for metrics. "
            f"Metrics will be exported to {cfg.kwargs['endpoint']}."
        )

        return otlp_metric_exporter

    @classmethod
    def _get_credentials(
        cls, cfg: EndpointConfig, insecure: bool
    ) -> Optional[grpc.ChannelCredentials]:
        return _get_credentials(cfg, insecure)


def configure_metrics(endpoints_file: str) -> None:
    """Configure metrics export for OTLP Collector.

    :param endpoints_file: The configuration file containing information about the
        metrics backend.
    :return: None.
    """
    cfg = read_endpoint_config(endpoints_file, ENDPOINTS_METRICS_KEY)

    if not cfg:
        logger.debug(
            "The OTLP Collector has not been configured to collect "
            "metrics. Skipping."
        )
        return None

    if cfg.type != ENDPOINTS_OTLP_BACKEND_TYPE:
        logger.warning(
            f"Unknown metrics backend type '{cfg.type}' "
            f"read from '{endpoints_file}', ignoring."
        )
        return None

    otlp_exporter = OTLPMetricConfigurer.configure_from_endpoint_config(cfg)
    metric_reader = PeriodicExportingMetricReader(otlp_exporter)
    set_meter_provider(
        MeterProvider(
            metric_readers=[metric_reader],
            resource=Resource.create(
                {
                    SERVICE_NAME: cfg.kwargs.get(
                        ENDPOINTS_TRACING_SERVICE_NAME_KEY, TRACING_SERVICE_NAME
                    )
                }
            ),
        )
    )

    MetricInstrumentProvider().register_instruments()


def _get_credentials(
    cfg: EndpointConfig, insecure: bool
) -> Optional[grpc.ChannelCredentials]:
    credentials = None
    if not insecure and ENDPOINTS_ROOT_CERTIFICATES_KEY in cfg.kwargs:
        with open(cfg.kwargs.get(ENDPOINTS_ROOT_CERTIFICATES_KEY), "rb") as f:
            root_cert = f.read()
        credentials = grpc.ssl_channel_credentials(root_certificates=root_cert)
    return credentials



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
import dataclasses
from enum import Enum
from functools import lru_cache
from typing import Any, Dict, List, Tuple, Type

import structlog
from rasa.shared.exceptions import RasaException

import rasa.shared.utils.common
from rasa.shared.utils.io import random_string


structlogger = structlog.get_logger()


def generate_stack_frame_id() -> str:
    """Generates a stack frame ID.

    Returns:
        The generated stack frame ID.
    """
    return random_string(8)


@lru_cache
def _get_all_subclasses() -> List[Type[DialogueStackFrame]]:
    stack_frame_subclasses = rasa.shared.utils.common.all_subclasses(DialogueStackFrame)

    # Get all the subclasses of DialogueStackFrame from the patterns package
    # in case these are not all imported at runtime
    modules = rasa.shared.utils.common.import_package_modules(
        "rasa.dialogue_understanding.patterns"
    )
    extra_subclasses = [
        clazz
        for module in modules
        for _, clazz in inspect.getmembers(module, inspect.isclass)
        if issubclass(clazz, DialogueStackFrame) and clazz not in stack_frame_subclasses
    ]

    stack_frame_subclasses.extend(extra_subclasses)
    return stack_frame_subclasses


class InvalidStackFrameType(RasaException):
    """Raised if a stack frame type is invalid."""

    def __init__(self, frame_type: str) -> None:
        """Creates a `InvalidStackFrameType`.

        Args:
            frame_type: The invalid frame type.
        """
        super().__init__(f"Invalid stack frame type '{frame_type}'.")


@dataclass
class DialogueStackFrame:
    """Represents the current flow step."""

    frame_id: str = field(default_factory=generate_stack_frame_id)
    """The ID of the current frame."""

    def as_dict(self) -> Dict[str, Any]:
        """Returns the `DialogueStackFrame` as a dictionary.

        Returns:
            The `DialogueStackFrame` as a dictionary.
        """

        def custom_asdict_factory(fields: List[Tuple[str, Any]]) -> Dict[str, Any]:
            """Converts enum values to their value."""

            def rename_internal(field_name: str) -> str:
                return field_name[:-1] if field_name.endswith("_") else field_name

            return {
                rename_internal(field): value.value
                if isinstance(value, Enum)
                else value
                for field, value in fields
            }

        data = dataclasses.asdict(self, dict_factory=custom_asdict_factory)
        data["type"] = self.type()
        return data

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        raise NotImplementedError

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> DialogueStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        raise NotImplementedError

    def context_as_dict(
        self, underlying_frames: List[DialogueStackFrame]
    ) -> Dict[str, Any]:
        """Returns the context of the frame."""
        return self.as_dict()

    @staticmethod
    def create_typed_frame(data: Dict[str, Any]) -> DialogueStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        typ = data.get("type")

        for clazz in _get_all_subclasses():
            try:
                if typ == clazz.type():
                    return clazz.from_dict(data)
            except NotImplementedError:
                # we don't want to raise an error if the frame type is not
                # implemented, as this is ok to be raised by an abstract class
                pass
        else:
            structlogger.warning("dialogue_stack.frame.unknown_type", data=data)
            raise InvalidStackFrameType(typ)



from __future__ import annotations
from typing import Dict, Text, Any, List

from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain
from rasa.shared.core.generator import TrackerWithCachedStates, TrainingDataGenerator
from rasa.shared.core.training_data.structures import StoryGraph


class TrainingTrackerProvider(GraphComponent):
    """Provides training trackers to policies based on training stories."""

    def __init__(self, config: Dict[Text, Any]) -> None:
        """Creates provider from config."""
        self._config = config

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns default configuration (see parent class for full docstring)."""
        return {
            "remove_duplicates": True,
            "unique_last_num_states": None,
            "augmentation_factor": 50,
            "tracker_limit": None,
            "use_story_concatenation": True,
            "debug_plots": False,
        }

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> TrainingTrackerProvider:
        """Creates component (see parent class for full docstring)."""
        return cls(config)

    def provide(
        self, story_graph: StoryGraph, domain: Domain
    ) -> List[TrackerWithCachedStates]:
        """Generates the training trackers from the training data.

        Args:
            story_graph: The story graph containing the test stories and rules.
            domain: The domain of the model.

        Returns:
            The trackers which can be used to train dialogue policies.
        """
        generator = TrainingDataGenerator(story_graph, domain, **self._config)
        return generator.generate()



from __future__ import annotations
import logging
import re
from typing import Any, Dict, List, Optional, Text

from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.storage import ModelStorage
from rasa.engine.storage.resource import Resource
import rasa.shared.utils.io
import rasa.nlu.utils.pattern_utils as pattern_utils
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_VALUE,
    ENTITY_ATTRIBUTE_START,
    ENTITY_ATTRIBUTE_END,
    TEXT,
    ENTITY_ATTRIBUTE_TYPE,
)
from rasa.nlu.extractors.extractor import EntityExtractorMixin

logger = logging.getLogger(__name__)


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.ENTITY_EXTRACTOR, is_trainable=True
)
class RegexEntityExtractor(GraphComponent, EntityExtractorMixin):
    """Extracts entities via lookup tables and regexes defined in the training data."""

    REGEX_FILE_NAME = "regex.json"

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            # text will be processed with case insensitive as default
            "case_sensitive": False,
            # use lookup tables to extract entities
            "use_lookup_tables": True,
            # use regexes to extract entities
            "use_regexes": True,
            # use match word boundaries for lookup table
            "use_word_boundaries": True,
        }

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> RegexEntityExtractor:
        """Creates a new `GraphComponent`.

        Args:
            config: This config overrides the `default_config`.
            model_storage: Storage which graph components can use to persist and load
                themselves.
            resource: Resource locator for this component which can be used to persist
                and load itself from the `model_storage`.
            execution_context: Information about the current graph run. Unused.

        Returns: An instantiated `GraphComponent`.
        """
        return cls(config, model_storage, resource)

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        patterns: Optional[List[Dict[Text, Text]]] = None,
    ) -> None:
        """Creates a new instance.

        Args:
            config: The configuration.
            model_storage: Storage which graph components can use to persist and load
                themselves.
            resource: Resource locator for this component which can be used to persist
                and load itself from the `model_storage`.
            patterns: a list of patterns
        """
        # graph component
        self._config = {**self.get_default_config(), **config}
        self._model_storage = model_storage
        self._resource = resource
        # extractor
        self.case_sensitive = self._config["case_sensitive"]
        self.patterns = patterns or []

    def train(self, training_data: TrainingData) -> Resource:
        """Extract patterns from the training data.

        Args:
            training_data: the training data
        """
        self.patterns = pattern_utils.extract_patterns(
            training_data,
            use_lookup_tables=self._config["use_lookup_tables"],
            use_regexes=self._config["use_regexes"],
            use_only_entities=True,
            use_word_boundaries=self._config["use_word_boundaries"],
        )

        if not self.patterns:
            rasa.shared.utils.io.raise_warning(
                "No lookup tables or regexes defined in the training data that have "
                "a name equal to any entity in the training data. In order for this "
                "component to work you need to define valid lookup tables or regexes "
                "in the training data."
            )
        self.persist()
        return self._resource

    def process(self, messages: List[Message]) -> List[Message]:
        """Extracts entities from messages and appends them to the attribute.

        If no patterns where found during training, then the given messages will not
        be modified. In particular, if no `ENTITIES` attribute exists yet, then
        it will *not* be created.

        If no pattern can be found in the given message, then no entities will be
        added to any existing list of entities. However, if no `ENTITIES` attribute
        exists yet, then an `ENTITIES` attribute will be created.

        Returns:
           the given list of messages that have been modified
        """
        if not self.patterns:
            rasa.shared.utils.io.raise_warning(
                f"The {self.__class__.__name__} has not been "
                f"trained properly yet. "
                f"Continuing without extracting entities via this extractor."
            )
            return messages

        for message in messages:
            extracted_entities = self._extract_entities(message)
            extracted_entities = self.add_extractor_name(extracted_entities)
            message.set(
                ENTITIES,
                message.get(ENTITIES, []) + extracted_entities,
                add_to_output=True,
            )
        return messages

    def _extract_entities(self, message: Message) -> List[Dict[Text, Any]]:
        """Extract entities of the given type from the given user message.

        Args:
            message: a message
        Returns:
            a list of dictionaries describing the entities
        """
        entities = []

        flags = 0  # default flag
        if not self.case_sensitive:
            flags = re.IGNORECASE

        for pattern in self.patterns:
            matches = re.finditer(pattern["pattern"], message.get(TEXT), flags=flags)

            for match in matches:
                start_index = match.start()
                end_index = match.end()
                entities.append(
                    {
                        ENTITY_ATTRIBUTE_TYPE: pattern["name"],
                        ENTITY_ATTRIBUTE_START: start_index,
                        ENTITY_ATTRIBUTE_END: end_index,
                        ENTITY_ATTRIBUTE_VALUE: message.get(TEXT)[
                            start_index:end_index
                        ],
                    }
                )

        return entities

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> RegexEntityExtractor:
        """Loads trained component (see parent class for full docstring)."""
        try:
            with model_storage.read_from(resource) as model_path:
                regex_file = model_path / cls.REGEX_FILE_NAME
                patterns = rasa.shared.utils.io.read_json_file(regex_file)
                return cls(
                    config,
                    model_storage=model_storage,
                    resource=resource,
                    patterns=patterns,
                )
        except (ValueError, FileNotFoundError):
            rasa.shared.utils.io.raise_warning(
                f"Failed to load {cls.__name__} from model storage. "
                f"This can happen if the model could not be trained because regexes "
                f"could not be extracted from the given training data - and hence "
                f"could not be persisted."
            )
            return cls(config, model_storage=model_storage, resource=resource)

    def persist(self) -> None:
        """Persist this model."""
        if not self.patterns:
            return
        with self._model_storage.write_to(self._resource) as model_path:
            regex_file = model_path / self.REGEX_FILE_NAME
            rasa.shared.utils.io.dump_obj_as_json_to_file(regex_file, self.patterns)
